:- include('clpFD.pl').


q:-     statistics(runtime,_),
        bridge(Ld,End), statistics(runtime,[_,Y]),
        write(Ld), nl,
        write(End), nl,
        write('time : '), write(Y), nl.

minimize(X,Y):-minimise(X,Y).
minimize_maximum(X,Y):-minimise_maximum(X,Y).

:-dynamic(recorded/3).
   
recorda(Key, Term, Ref):-
    g_read('$record_counter',R), NR=R-1, g_assign('$record_counter',NR),
    Ref='$ref'(NR, NR),
    asserta(recorded(Key, Term, Ref)).
recordz(Key, Term, Ref):-
    g_read('$record_counter',R), NR=R-1, g_assign('$record_counter',NR),
    Ref='$ref'(NR, NR),
    assertz(recorded(Key, Term, Ref)).
erase(Ref):-
    recorded(Key, Term, Ref), retract(recorded(Key, Term, Ref)).

% bridge.pl


bridge(Ld,End):-
	data_structure(Ld,Disj,End),
	constraint(Ld),
	minimize(disjunction(Disj),End).
%bridge(Ld,End):-
%	data_structure(Ld,Disj,End),
%	constraint(Ld),
%	disjunction(Disj),
%	minimize(choice(Disj,Ld),End).

bridge_and_redundant(Ld,End):-
	data_structure_and_redundant(Ld,Disj,End,Lv),
	constraint(Ld),
	minimize_maximum(choice(Disj,Ld),[End|Lv]).

/* Create the data structures */

data_structure(DataList,Disj,End):-
	jobs(Jobs),
	make_data(Jobs,DataList),
	ressources(Ressources),
	make_disj(Ressources,DataList,Disj),
	member(data(pe,_,End,_),DataList).

data_structure_and_redundant(DataList,Disj,End,Lv):-
	data_structure(DataList,Disj,End),
	redundant(Ressources,DataList,Lv).
	
% constructive disjunction on sta disj

disjunction([]).
disjunction([disj(Stai,Cij,Staj,Cji)|L]):-
	impose(or(Staj.>=Stai + Cij,Stai.>=Staj + Cji)),
	disjunction(L).


/* Put the constraints */

constraint(Ld):-
	precedences(PrecedenceList),
	precedence_constraints(PrecedenceList,Ld),
	distances(DistanceList),
	distance_constraints(DistanceList,Ld).

precedence_constraints([],_).
precedence_constraints([[T1,T2]|T],DataList):-
	precedence_constraint(T1,T2,DataList),
	precedence_constraints(T,DataList).

precedence_constraint(T1,T2,DataList):-
	member(data(T1,D1,S1,_),DataList),
	member(data(T2,_,S2,_),DataList),
	S2.>=S1+D1.

distance_constraints([],_).
distance_constraints([H|T],Data):-
	distance_constraint(H,Data),
	distance_constraints(T,Data).

/*
supdis(4,e(b1),e(s1)) is 4>=b1 + db1 - (s1 + ds1)
infdis(e(b1),e(s1),3) is b1 + db1 - (s1 + ds1) >=3
*/

distance_constraint(supdis(D,s(T1),s(T2)),Data):-!,
	member(data(T1,_,S1,_),Data),
	member(data(T2,_,S2,_),Data),
	D+S2.>=S1.
distance_constraint(supdis(D,s(T1),e(T2)),Data):-!,
	member(data(T1,_,S1,_),Data),
	member(data(T2,D2,S2,_),Data),
	D+S2+D2.>=S1.
distance_constraint(supdis(D,e(T1),s(T2)),Data):-!,
	member(data(T1,D1,S1,_),Data),
	member(data(T2,_,S2,_),Data),
	D+S2.>=S1+D1.
distance_constraint(supdis(D,e(T1),e(T2)),Data):-!,
	member(data(T1,D1,S1,_),Data),
	member(data(T2,D2,S2,_),Data),
	D+S2+D2.>=S1+D1.

distance_constraint(infdis(s(T1),s(T2),D),Data):-!,
	member(data(T1,_,S1,_),Data),
	member(data(T2,_,S2,_),Data),
	S1.>=D+S2.
distance_constraint(infdis(e(T1),s(T2),D),Data):-!,
	member(data(T1,D1,S1,_),Data),
	member(data(T2,_,S2,_),Data),
	S1+D1.>=D+S2.
distance_constraint(infdis(s(T1),e(T2),D),Data):-!,
	member(data(T1,_,S1,_),Data),
	member(data(T2,D2,S2,_),Data),
	S1.>=D+S2+D2.
distance_constraint(infdis(e(T1),e(T2),D),Data):-!,
	member(data(T1,D1,S1,_),Data),
	member(data(T2,D2,S2,_),Data),
	S1+D1.>=D+S2+D2.

/* Make choice on disjunctive and enumerate */

choice(Disj,Ld) :-
	clean_nbb,
	recorda(nbb,nbb(0),_),
	make_choice(Disj),
	enumerate_on(Ld,Ls),
	labeling(Ls),
	write('Solution : '),write(Ls),nl,
	!,
	write_backtrack.
choice(_,_) :-
	write_backtrack,
	fail.

enumerate_on([],[]).
enumerate_on([data(_,_,S,_)|T],[S|T1]):-enumerate_on(T,T1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Labeling Disj Normal
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_choice([]).
make_choice([disj(Stai,Dij,Staj,Dji)|T]) :-
	disjunctive(Stai,Dij,Staj,Dji),
	make_choice(T).

%disjunctive as a choice point
disjunctive(Si,Di,Sj, Dj) :-
	minmaxvalue(Si, Mini, Maxi),
	minmaxvalue(Sj, Minj, Maxj),
	(Maxj < Mini + Di;Maxi < Minj + Dj),
	!.
disjunctive(Stai,Dij,Staj,_Dji) :-
	Staj.>=Stai + Dij.
disjunctive(Stai,_Dij,Staj,Dji) :-
	inc,
	Stai.>=Staj + Dji.

/*
make_disj(ResourceList,DataStructureList,DisjunctiveList).
*/

make_disj([],_,[]).
make_disj([[_,Tasks]|T],Data,L):-
	make_one_disj(Tasks,Data,L1),
	make_disj(T,Data,L2),
	append(L1,L2,L).

make_one_disj([],_,[]).
make_one_disj([H|T],Data,L):-
	make_oneXone_disj(H,T,Data,L1),
	make_one_disj(T,Data,L2),
	append(L1,L2,L).

make_oneXone_disj(_,[],_,[]).
make_oneXone_disj(H,[H1|T1],Data,[H2|T2]):-
	member(data(H,D1,B1,_),Data),
	member(data(H1,D2,B2,_),Data),
	H2=disj(B1,D1,B2,D2),
	make_oneXone_disj(H,T1,Data,T2).

redundant([],_,[]).
redundant([H|T],Ld,[H1|T1]):-one_redundant(H,Ld,H1),redundant(T,Ld,T1).

one_redundant([_,L],Ld,AtLeast):-
	start_time_duration(L,Ld,SList,DList),
	sum_list(DList,D),
	domain(AtLeast,[0,200]),
	domain(Min,[0,200]),
	min_list(SList,Min),
	eqln(AtLeast,Min+D).

start_time_duration([],_,[],[]).
start_time_duration([H|T],Ld,[S1|T1],[D1|T2]):-
	member(data(H,D1,S1,_),Ld),
	start_time_duration(T,Ld,T1,T2).


/*
make_data(TaskList,DataList)
*/

make_data([],[]).
make_data([H|T],[HD|TD]):-
	make_one_data(H,HD),
	make_data(T,TD).

make_one_data(H,data(Task,Duration,Start,_)):-
	Task=H,
	jobs(JOBS),
	durations(DURATIONS),
	nth(N,JOBS,H),
	nth(N,DURATIONS,Duration),
%	totalduration(DURATIONS,TotalDuration),
%	domain(Start,[0,TotalDuration]),
%	domain(End,[0,TotalDuration]),
	domain(Start,[0,200]).

totalduration([],0).
totalduration([H|T],N):-totalduration(T,N1),N is N1 + H.

/* Additional constraints */

/* Data for the bridge scheduling problem */

jobs([pa,a1,a2,a3,a4,a5,a6,p1,p2,ue,s1,s2,s3,s4,s5,s6,b1,b2,b3,b4,b5,b6,ab1,
    ab2,ab3,ab4,ab5,ab6,m1,m2,m3,m4,m5,m6,l,t1,t2,t3,t4,t5,ua,v1,v2,pe]).

durations([0,4,2,2,2,2,5,20,13,10,8,4,4,4,4,10,1,1,1,1,1,1,1,1,1,1,1,1,16,8,
    8,8,8,20,2,12,12,12,12,12,10,15,10,0]).

precedences([
[pa,p1],
[pa,p2],
[pa,a1],
[pa,a2],
[pa,a5],
[pa,a6],
[pa,ue],
[pa,l],

[p1,a3],
[p2,a4],

[a1,s1],
[a2,s2],
[a3,s3],
[a4,s4],
[a5,s5],
[a6,s6],

[s1,b1],
[s2,b2],
[s3,b3],
[s4,b4],
[s5,b5],
[s6,b6],

[b1,ab1],
[b2,ab2],
[b3,ab3],
[b4,ab4],
[b5,ab5],
[b6,ab6],

[ab1,m1],
[ab2,m2],
[ab3,m3],
[ab4,m4],
[ab5,m5],
[ab6,m6],

[m1,t1],
[m2,t1],
[m2,t2],
[m3,t2],
[m3,t3],
[m4,t3],
[m4,t4],
[m5,t4],
[m5,t5],
[m6,t5],

[l,t1],
[l,t2],
[l,t3],
[l,t4],
[l,t5],

[ue,ua],
[ua,pe],

[m1,v1],
[t1,v1],
[m6,v2],
[t5,v2],

[v1,pe],
[t1,pe],
[t2,pe],
[t3,pe],
[t4,pe],
[t5,pe],
[v2,pe]
]).

ressources([
[pile_driver,[p1,p2]],
[crane,[l,t1,t2,t3,t5,t4]],
[bricklaying,[m1,m2,m3,m4,m5,m6]],
[carpentary,[s1,s2,s3,s4,s5,s6]],
[excavator,[a1,a2,a3,a4,a5,a6]],
[caterpillar,[v1,v2]],
[concrete_mixer,[b1,b2,b3,b4,b5,b6]]
]).

/*
supdis(4,e(b1),e(s1)) is 4>=b1 + db1 - (s1 + ds1)
infdis(e(b1),e(s1),3) is b1 + db1 - (s1 + ds1) >=3
*/

distances([


supdis(4,e(b1),e(s1)),
supdis(4,e(b2),e(s2)),
supdis(4,e(b3),e(s3)),
supdis(4,e(b4),e(s4)),
supdis(4,e(b5),e(s5)),
supdis(4,e(b6),e(s6)),

supdis(3,s(s1),e(a1)),
supdis(3,s(s2),e(a2)),
supdis(3,s(s3),e(a3)),
supdis(3,s(s4),e(a4)),
supdis(3,s(s5),e(a5)),
supdis(3,s(s6),e(a6)),

supdis(3,s(s3),e(p1)),
supdis(3,s(s4),e(p2)),

infdis(s(s1),s(ue),6),
infdis(s(s2),s(ue),6),
infdis(s(s3),s(ue),6),
infdis(s(s4),s(ue),6),
infdis(s(s5),s(ue),6),
infdis(s(s6),s(ue),6),

infdis(s(ua),e(m1),-2),
infdis(s(ua),e(m2),-2),
infdis(s(ua),e(m3),-2),
infdis(s(ua),e(m4),-2),
infdis(s(ua),e(m5),-2),
infdis(s(ua),e(m6),-2),

supdis(30,s(l),s(pa)),
infdis(s(l),s(pa),30),

supdis(0,s(ab1),e(b1)),
supdis(0,s(ab2),e(b2)),
supdis(0,s(ab3),e(b3)),
supdis(0,s(ab4),e(b4)),
supdis(0,s(ab5),e(b5)),
supdis(0,s(ab6),e(b6)),

infdis(s(ab1),e(b1),0),
infdis(s(ab2),e(b2),0),
infdis(s(ab3),e(b3),0),
infdis(s(ab4),e(b4),0),
infdis(s(ab5),e(b5),0),
infdis(s(ab6),e(b6),0)

]).


%%%%%%%%%%%% 
% tools



inc :-
	recorded(nbb,nbb(N),Ref),
	N1 is N+1,
	erase(Ref),
	recorda(nbb,nbb(N1),_).
	
clean_nbb :-
	recorded(nbb,_,Ref), erase(Ref),fail;true.

write_backtrack :-
	recorded(nbb,nbb(Nb),_Ref),
	write('number of backtrack : '), 
	write(Nb),nl.
	



