/*-------------------------------------------------------------------------* 
 * GNU Prolog                                                              * 
 *                                                                         * 
 * Part  : Prolog attributes engine                                        * 
 * File  : attributes.pl                                                   * 
 * Descr.: attributes manipulation tools - prolog file                     * 
 * Author: Remy Haemmerle                                                  * 
 *                                                                         * 
 * Copyright (C) 2002-2003 INRIA, Remy Haemmerle                           * 
 *                                                                         * 
 * GNU Prolog is free software; you can redistribute it and/or modify it   * 
 * under the terms of the GNU General Public License as published by the   * 
 * Free Software Foundation; either version 2, or any later version.       * 
 *                                                                         * 
 * GNU Prolog is distributed in the hope that it will be useful, but       * 
 * WITHOUT ANY WARRANTY; without even the implied warranty of              * 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        * 
 * General Public License for more details.                                * 
 *                                                                         * 
 * You should have received a copy of the GNU General Public License along * 
 * with this program; if not, write to the Free Software Foundation, Inc.  * 
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     * 
 *-------------------------------------------------------------------------*/

:-built_in.

'$use_attributes'.

attributed(X):-
    '$call_c_test'(attributed(X)).

'$put_atts'(Var, I, Att):-
     '$call_c_test'(put_attributes(Var,I,Att)).

'$get_atts'(Var, I, Att):-
     '$call_c_test'(get_attributes(Var, I, Att)).

put_atts(Var, _):-
    (non_generic_var(Var); fd_var(Var)),!, 
    '$pl_err_type'(variable, Var).

put_atts(Var,Att):-
    set_bip_name(put_atts,2),
    generic_var(Att) -> 
        '$pl_err_instantiation'
    ;
         (functor(Att,F,N),
	 '$get_attr_info'(F,N,I) ->
	     (N == 1 ->
	         arg(1,Att,A);
                 A = Att), 
	     '$call_c_test'(put_attributes(Var,I,A))
        ;
             '$pl_err_domain'(attributes, F/N)).

get_atts(Var,Att):-
    set_bip_name(get_atts,2),
    generic_var(Att) -> 
        '$pl_err_instantiation'
    ;
         (functor(Att,F,N),
	 '$get_attr_info'(F,N,I) ->
	     (N == 1 ->
	         arg(1,Att,A);
                 A = Att), 
             '$call_c_test'(get_attributes(Var,I,A))
        ;
             '$pl_err_domain'(attributes, F/N)).

force_binding(X,Y):-
    '$call_c_test'(force_binding(X,Y)).


'$add_attributes'([Functor/Arity|L]):-
    nonvar(L), '$add_attributes'(L),
    '$call_c'(get_nb_atts(Index)),
    '$call_c_test'(inc_nb_atts) ->
        asserta('$get_attr_info'(Functor,Arity,Index));
        format('warning : too many attributes defined : attribute ~q not added\n', [Functor/Arity]).
'$add_attributes'([]).

/*
'$add_attr_predicate'([Functor]):-
    '$call_c'(add_verify_atts_predicate(Functor)).
*/

'$add_attr_verify_predicate'([Functor]):-
    g_read('$verify_attributes_predicates', Functors),
    list(Functors) ->
        g_assign('$verify_attributes_predicates',([Functor|Functors]));
        g_assign('$verify_attributes_predicates',([Functor])).

'$add_attr_portray_predicate'([Functor]):-
    g_read('$attribute_portray_predicates',(Functors)), 
    list(Functors) -> 
        g_assign('$attribute_portray_predicates',([Functor|Functors]));
        g_assign('$attribute_portray_predicates',([Functor])).

:-dynamic('$get_attr_info'/3).

'$get_attr_info'(Functor,Arity,Index):-
    '$get_static_attr_info'(Functor,Arity,Index).

%portray_attributes(Output, QueryVars):-write((Output, QueryVars)), nl.

portray_attributes(Output, QueryVarsIn):- 
    term_variables(QueryVarsIn, QueryVars),
    filter_attributes(QueryVars, AttVars),% sort(AttVars),
%    attributes_projetction_list(Output, AttVars),
    portray_attributes_list(Output, QueryVarsIn, AttVars), fail;
    true.

filter_attributes([H|T],L):-!,
    (attributed(H) -> L = [H|T2]; L = T2),
    filter_attributes(T,T2).
filter_attributes([],[]). 

/*
filter_attributes([Name=Var|T], [Var|T1]):-
    attributed(Var),!, filter_attributes(T,T1).
filter_attributes([_|T], T1):-
    !, filter_attributes(T,T1).
filter_attributes([],[]). 
*/

portray_attributes_list(Output, QueryVars, AttVars):-
    g_read('$attribute_portray_predicates',(Functors)),
    portray_attributes_list(Output, QueryVars, AttVars, Functors).


/*
portray_attributes_list(Output, QueryVars, AttVars, [H|T]):-!,
    portray_attributes_list(Output, QueryVars, AttVars, T),
    portray_attributes_list_b(Output, QueryVars, AttVars, H).

portray_attributes_list(_,_,_,[]).


portray_attributes_list_b(Output, QueryVars, [Var|T], F):-(
    (call_with_args(F, Var, G),
    '$set_query_name'(QueryVars),
    nl, write_term(G, [space_args(true)], max_depth(16)]),
    fail;true),
    portray_attributes_list_b(Output, QueryVars, T, F)).
portray_attributes_list_b(_, [], _).
*/


portray_attributes_list(Output, QueryVars, AttVars, [H|T]):- !,
    (call_with_args(H, AttVars, Constraints),
    '$set_query_name'(QueryVars),
    write_constraints(Constraints);
    true),
    portray_attributes_list(Output, QueryVars, AttVars, T).
/*
portray_attributes_list(Output, QueryVars, AttVars, []):-
   '$portray_frozen'(AttVars, Constraints),
   '$set_query_name'(QueryVars),
    write_constraints(Constraints).
*/
portray_attributes_list(_, _, _, []).

write_constraints([H|T]):-
    (member_ptr(H,T) ->
      true;
      nl, write_term(H, [space_args(true)])),
    write_constraints(T).


'$set_query_name'([Name=Var|T]):- 
     (attributed(Var),!, force_binding(Var,'$VARNAME'(Name)); 
      var(Var), !, Var='$VARNAME'(Name)), 
    '$set_query_name'(T).
'$set_query_name'([_|T]):-!,'$set_query_name'(T).
'$set_query_name'([]).


member_ptr(X,[H|_]):-X==H.
member_ptr(X,[_|T]):-member_ptr(X,T).
