/*-------------------------------------------------------------------------* 
 * GNU Prolog                                                              * 
 *                                                                         * 
 * Part  : Prolog attributes engine                                        * 
 * File  : attributes_c.c                                                  * 
 * Descr.: attributes manipulatiopn tools                                  * 
 * Author: Remy Haemmerle                                                  * 
 *                                                                         * 
 * Copyright (C) 2002 INRIA, Remy Haemmerle                                * 
 *                                                                         * 
 * GNU Prolog is free software; you can redistribute it and/or modify it   * 
 * under the terms of the GNU General Public License as published by the   * 
 * Free Software Foundation; either version 2, or any later version.       * 
 *                                                                         * 
 * GNU Prolog is distributed in the hope that it will be useful, but       * 
 * WITHOUT ANY WARRANTY; without even the implied warranty of              * 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        * 
 * General Public License for more details.                                * 
 *                                                                         * 
 * You should have received a copy of the GNU General Public License along * 
 * with this program; if not, write to the Free Software Foundation, Inc.  * 
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     * 
 *-------------------------------------------------------------------------*/

#include "engine_pl.h"
#include "engine_atts.h"

/*---------------------------------*
 * Global Variables                *
 *---------------------------------*/

WamWord n_attributes = NB_STATIC_ATTRIBUTES;
WamWord verify_atts_list = NIL_WORD;

Bool 
get_nb_atts(WamWord n)
{
  // more efficient than Unify(n, n_attributes);
  return Get_Integer_Tagged(Tag_INT(n_attributes), n);
}

Bool 
inc_nb_atts()
{
  if (n_attributes >= MAX_NB_ATTS)
    return FALSE;
  else
    {
      n_attributes ++;
      return TRUE;
    }
}
/*
void
add_verify_atts_predicate(WamWord Functor)
{
  WamWord * tmp;
  WamWord word, tag_mask;


  tmp = malloc((OFFSET_CAR+2) * sizeof(WamWord));
  DEREF(Functor, word, tag_mask);
  Car(tmp) = word;
  Cdr(tmp) = verify_atts_list;
  verify_atts_list = Tag_LST(tmp);
}

Bool
get_verify_atts_predicates(WamWord Functors)
{
  return Unify(Functors, verify_atts_list);
}
*/
Bool
attributed(WamWord s_word)
{
  WamWord word, tag_mask;
  
  DEREF(s_word, word, tag_mask);

  return (tag_mask == TAG_ATT_MASK);
}

Bool
put_attributes(WamWord s_word, WamWord Index, WamWord Attr)
{
  WamWord word, tag_mask;
  WamWord * addr;

  DEREF(Index, word, tag_mask);
  if (tag_mask != TAG_INT_MASK) return FALSE;
  Index = UnTag_INT(word);

  DEREF(s_word, word, tag_mask);

   if (tag_mask == TAG_REF_MASK)
     {
       Bind_UV(UnTag_REF(word), Make_Self_Ref(H));
       Global_Push(Tag_ATT(H));
       Global_Push(1 << Index);
       
       DEREF(Attr, word, tag_mask);
       H[Index] = word;
       
       H += n_attributes;
     }
   else if (tag_mask == TAG_ATT_MASK)
     {
       addr = UnTag_ATT(word);
       if (ACTIVE_ATTRIBUTE(addr[1], Index))
	 {
	   DEREF(Attr, word, tag_mask);
	   Bind_OV(addr+2+Index, word);  
	 }
       else
	 {
	   Bind_OV(addr+1, addr[1] | (1 << Index));
	   DEREF(Attr, word, tag_mask);
   	   addr[2+Index]=word;
	 }
     }
   else return FALSE;

   return TRUE;
}
     
Bool
get_attributes(WamWord s_word, WamWord Index, WamWord Attr)
{
  WamWord word, tag_mask;
  WamWord * addr;

  DEREF(s_word, word, tag_mask);

  if (tag_mask == TAG_ATT_MASK)
    {
      addr = UnTag_ATT(word);
  
      DEREF(Index, word, tag_mask);
      if (tag_mask != TAG_INT_MASK)
	return FALSE;
      Index = UnTag_INT(word);
      
      if (ACTIVE_ATTRIBUTE(addr[1], Index))
	return Unify(addr[Index +2], Attr);
    }

  return FALSE;
}
       
Bool 
force_binding(WamWord s_word, WamWord other)
{
  WamWord u_word,  tag_mask;
  WamWord v_word;
    
  DEREF(s_word, u_word, tag_mask);
  if (tag_mask == TAG_ATT_MASK) 
    {
      Bind_OV(UnTag_ATT(u_word), other);
      return TRUE;
    }
  
  DEREF(other, v_word, tag_mask);
  
  return u_word == v_word;
}
