/*-------------------------------------------------------------------------* 
 * GNU Prolog  RH                                                          * 
 *                                                                         * 
 * Part  : Prolog attributes engine                                        * 
 * File  : dif.pl                                                          * 
 * Descr.: dif constrant                                                   * 
 * Author: Remy Haemmerle                                                  * 
 *                                                                         * 
 * Copyright (C) 2004 INRIA, Remy Haemmerle                                * 
 *                                                                         * 
 * GNU Prolog is free software; you can redistribute it and/or modify it   * 
 * under the terms of the GNU General Public License as published by the   * 
 * Free Software Foundation; either version 2, or any later version.       * 
 *                                                                         * 
 * GNU Prolog is distributed in the hope that it will be useful, but       * 
 * WITHOUT ANY WARRANTY; without even the implied warranty of              * 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        * 
 * General Public License for more details.                                * 
 *                                                                         * 
 * You should have received a copy of the GNU General Public License along * 
 * with this program; if not, write to the Free Software Foundation, Inc.  * 
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     * 
 *-------------------------------------------------------------------------*/


'$use_dif'.
 
%:-attribute(dif/1).
:-verify_attributes_predicate('$verify_dif').
:-portray_attributes_predicate('$portray_dif').


dif(X, Y):-
    '$dif'([X], [Y], dif(X, Y)).


'$dif'([H1|T1], [H2|T2], Portray):-
    ( generic_var(H1) ->
        ( generic_var(H2) ->  
            ( H1 == H2 -> 
                '$dif'(T1, T2, Portray)
            ;  
                '$add_dif_var'(H1, ['$dif'(H2, T1, T2, Portray, _)]),
                '$add_dif_var'(H2, [])
	    )
        ;
            '$add_dif_const'(H1, ['$dif'(H2, T1, T2, Portray, _)])
	)
    ; 
        ( generic_var(H2) ->
            '$add_dif_const'(H2, ['$dif'(H1, T1, T2, Portray, _)])     
        ;
            ( functor(H1, F, N), functor(H2, F, N) ->
               (
	           N=0, H1 \== H2, !
	       ;
                   H1 =.. [_| L1], append(L1, T1, NL1),
                   H2 =.. [_| L2], append(L2, T2, NL2),
                   '$dif'(NL1, NL2,  Portray)	  
	       )
            ;
               true
	    )
	)
    ).
        


'$verify_dif'(X, Y, G):-
    ( get_atts(X, '$dif'(LX)) ->
        ( generic_var(Y) ->
            ( get_atts(Y, '$dif'(LY)) ->
                '$verify_dif_var'(Y, LX, G1),
	        '$verify_dif_var'(X, LY, G2),
                G = (G1, G2)
            ;   
                G = true
            ),
            '$add_dif'(Y, LX)
        ;
            '$verify_dif_const'(Y, LX, G)
        )
    ;
        G = true
    ).


'$verify_dif_var'(_, [], true):-!.
'$verify_dif_var'(Y, ['$dif'(X, L1, L2, Portray, Flag) | T], G):-
    ( var(Flag), Y == X ->
        Flag = true, G = ('$dif'(L1, L2, Portray), G1)
    ;
        G = G1
    ),
    '$verify_dif_var'(Y, T, G1).


'$verify_dif_const'(_, [], true):-!.
'$verify_dif_const'(Y, ['$dif'(X, L1, L2, Portray, Flag) | T], G):- 
    ( nonvar(Flag) ->
        G = G1
    ;
        ( generic_var(X) ->
            '$add_dif_const'(X, ['$dif'(Y, L1, L2, Portray, Flag)]),
            G = G1
        ;
	    Flag = true,
            ( functor(X, F, N), functor(Y, F, N) ->
                (
		   N=0, X \== Y, G = G1, !
	        ;
	           X =.. [_| T1], append(T1, L1, NL1),
	           Y =.. [_| T2], append(T2, L2, NL2),
	           G = ('$dif'(NL1, NL2,  Portray), G1)	  
	        )
            ;
	        G = G1

            )
        )
    ),
    '$verify_dif_const'(Y, T, G1).

    
'$add_dif_var'(X, L):-
    '$add_dif'(X, L).

'$add_dif_const'(X, L):-
    '$add_dif'(X, L).


'$add_dif'(X, L):-
    get_atts(X, '$dif'(OL)) ->
      append(L, OL, NL),
      put_atts(X, '$dif'(NL))
    ;
      put_atts(X, '$dif'(L)).


'$portray_dif'([], []):-!.
'$portray_dif'([H|T], L):-
    '$portray_dif'(T, L1),
    ( get_atts(H, '$dif'(L2)) ->
        '$portray_dif2'(L2, L3),
         append(L3, L1, L)
    ; 
         L1 = L
    ).
    

'$portray_dif2'([],[]):-!.
'$portray_dif2'(['$dif'(_, _, _, Portray, Flag)|T] , L):-
    ( var(Flag) ->
        Flag = true, 
        L = [Portray|L1]
      ;
        L = L1
    ),
    '$portray_dif2'(T, L1).






