/*-------------------------------------------------------------------------* 
 * GNU Prolog                                                              * 
 *                                                                         * 
 * Part  : Prolog attributes engine                                        * 
 * File  : froze.pl                                                        * 
 * Descr.: freeze coroutines                                               * 
 * Author: Remy Haemmerle                                                  * 
 *                                                                         * 
 * Copyright (C) 2002 INRIA, Remy Haemmerle                                * 
 *                                                                         * 
 * GNU Prolog is free software; you can redistribute it and/or modify it   * 
 * under the terms of the GNU General Public License as published by the   * 
 * Free Software Foundation; either version 2, or any later version.       * 
 *                                                                         * 
 * GNU Prolog is distributed in the hope that it will be useful, but       * 
 * WITHOUT ANY WARRANTY; without even the implied warranty of              * 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        * 
 * General Public License for more details.                                * 
 *                                                                         * 
 * You should have received a copy of the GNU General Public License along * 
 * with this program; if not, write to the Free Software Foundation, Inc.  * 
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     * 
 *-------------------------------------------------------------------------*/

%:-attribute('$frozen'/1).
%:-verify_attributes_predicate('$verif_frozen').
:-portray_attributes_predicate('$portray_frozen').

'$use_freeze'.

freeze(X,G):-
    non_generic_var(X) -> call(G);
    get_atts(X,'$frozen'(G2)) ->
        put_atts(X, '$frozen'((G2,G)));
        put_atts(X, '$frozen'((G))).

frozen(X,G):-get_atts(X, '$frozen'(G)),!.
frozen(X,true):-generic_var(X).


'$verif_frozen'(X,Y,G):- 
    (
      get_atts(X, '$frozen'(Gx)) ->
      (
	generic_var(Y) ->
	(
	  get_atts(Y, '$frozen'(Gy)) -> 
	  put_atts(Y, '$frozen'((Gx,Gy)))
	;
	  put_atts(Y, '$frozen'(Gx))
	),
	G=true
      ;
%	    G=(format("wake up ~q~n", [Gx]), Gx, format("end ~q~n", [Gx])))
	G= Gx
      )
    ;
      G=true
    ).

/*
portray_frozen_list([H1|T1],[H2|T2]):-!,
    portray_frozen(H1,H2),
    portray_frozen_list(T1,T2).
    
portray_frozen_list([],[]).
*/

'$portray_frozen'([X|T],L):-
    (get_atts(X, '$frozen'(G1)) -> 
      catch((portray(G1,G),!; G = freeze(X,G1)),
	     error(existence_error(procedure,portray/2),_),
	     G = freeze(X,G1)),
      L = [G|T2];
      L = T2),
    '$portray_frozen'(T,T2).
'$portray_frozen'([],[]).
