/*-------------------------------------------------------------------------*
 * GNU Prolog                                                              *
 *                                                                         *
 * Part  : Attributes Engine                                               *
 * File  : tools_c.c                                                       *
 * Descr.: generic tools                                                   *
 * Author: Remy Haemmerle                                                  *
 *                                                                         *
 * Copyright (C) 2001-2002 Remy Haemmerle                                  *
 *                                                                         *
 * GNU Prolog is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2, or any later version.       *
 *                                                                         *
 * GNU Prolog is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc.  *
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     *
 *-------------------------------------------------------------------------*/

#include "engine_pl.h"
#include "tools.h"

/*-------------------------------------------------------------------------*/
/* GROUND                                                                  */
/*                                                                         */
/* This functions test wheter X is completely instantiated, i.e. free of   */
/* unbound variables. In this context, FD and CLP(R) variables and are     */
/* treated  as nonground.                                                  */
/*                                                                         */
/* Input:                                                                  */
/*    word  : term tested                                                  */
/*                                                                         */
/*-------------------------------------------------------------------------*/
Bool
Is_Ground(WamWord start_word)
{
  WamWord word, tag, *adr;
  int i,n;

  DEREF(start_word, word, tag);

  switch(tag)
    {
    case TAG_REF_MASK:
    case TAG_FDV_MASK:
    case TAG_ATT_MASK:
      return FALSE;
    case TAG_LST_MASK:
      adr = UnTag_LST(word);
      return (Is_Ground(Car(adr)) && Is_Ground(Cdr(adr))); 
    case TAG_STC_MASK:
      adr=UnTag_STC(word);
      n = Arity(adr);
      for(i=0; i< n; i++) 
	if (! Is_Ground(Arg(adr,i) )) return FALSE;
    default: 
      return TRUE;
    }
}  
