/*-------------------------------------------------------------------------* 
 * GNU Prolog                                                              * 
 *                                                                         * 
 * Part  : Prolog attributes engine                                        * 
 * File  : wake_c.c                                                        * 
 * Descr.: wake engine                                                     * 
 * Author: Remy Haemmerle                                                  * 
 *                                                                         * 
 * Copyright (C) 2002 INRIA, Remy Haemmerle                                * 
 *                                                                         * 
 * GNU Prolog is free software; you can redistribute it and/or modify it   * 
 * under the terms of the GNU General Public License as published by the   * 
 * Free Software Foundation; either version 2, or any later version.       * 
 *                                                                         * 
 * GNU Prolog is distributed in the hope that it will be useful, but       * 
 * WITHOUT ANY WARRANTY; without even the implied warranty of              * 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        * 
 * General Public License for more details.                                * 
 *                                                                         * 
 * You should have received a copy of the GNU General Public License along * 
 * with this program; if not, write to the Free Software Foundation, Inc.  * 
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     * 
 *-------------------------------------------------------------------------*/

#include "engine_pl.h"
#include "engine_atts.h"


/*---------------------------------*
 * Global Variables                *
 *---------------------------------*/


void 
Wake(int n, CodePtr next_inst)
{
  int i;

#ifndef  M_ix86
  if (!TEST_WAKE_SIGNAL()) return;
#endif

  Allocate(n+3);
  Y(E,0) = (WamWord) n;
  Y(E,1) = (WamWord) next_inst;

  for (i=0; i<=n; i++)
    Y(E,i+2) = X(i);

  UNSET_WAKE_SIGNAL();
  
  CP=Prolog_Predicate(CONTINUE_AFTER_WAKE,0);
    
  Goto_Predicate(WAKE_DEAMON,0);
}


CodePtr 
Restore_Env()
{
  int i;
  int n;
  WamCont next_inst; 


  n = (int) Y(E,0);
  next_inst = (CodePtr) Y(E,1);

  for (i=0; i<=n; i++)
    X(i) = Y(E,i+2);

  Deallocate();
  return next_inst;
}

void 
Wake_Push(WamWord var, WamWord other)
{
  WamWord *cur_WT;

  SET_WAKE_SIGNAL();

  //  Trail_OV(&WT);
  cur_WT=WT++;
  *cur_WT=Tag_LST(WT);
  *WT++=var;
  cur_WT=WT++;
  *cur_WT=Tag_LST(WT);
  *WT++=other;
}

  
void Attributes_On()
{
  Bind_OV(&WAKE_STATUS, TRUE);
}

void Attributes_Off()
{
  Bind_OV(&WAKE_STATUS, FALSE);
}


Bool
Get_Wake_Stack(WamWord Stack)
{
  WamWord * tmp;

  tmp = WB;

  *WT++=NIL_WORD;
  //  Trail_OV(&WB);
  WB=WT;

  return Unify(*tmp, Stack);
}
