/*-------------------------------------------------------------------------*
 * GNU Prolog                                                              *
 *                                                                         *
 * Part  : CLP(R) Engine                                                   *
 * File  : clpr_c.c                                                        *
 * Descr.: CLP(R) solver interface                                         *
 * Author: Remy Haemmerle                                                  *
 *                                                                         *
 * Copyright (C) 2001-2002 INRIA, Remy Haemmerle                           *
 *                                                                         *
 * GNU Prolog is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2, or any later version.       *
 *                                                                         *
 * GNU Prolog is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc.  *
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     *
 *-------------------------------------------------------------------------*/

#include "clpr.h"

/*---------------------------------*
 * Constants                       *
 *---------------------------------*/


/*---------------------------------*/
/* Global Variables                */
/*---------------------------------*/

inclp * vpara = NULL;
WamWord CLPR_index_cmptr=1;
WamWord CLPR_level=0;

WamWord CLPR_var[CLPR_VAR_MAX];

WamWord clpr_too_big;

WamWord eq_atom;
WamWord le_atom;
int coma_symbol;


static int cmptr=-1;

/*---------------------------------*/
/* Function Prototypes             */
/*---------------------------------*/

void CLPR_back();
WamWord Mk_Tuple(WamWord, WamWord);



/*-------------------------------------------------------------------------*
 * CLPR_BACK                                                               *
 *                                                                         *
 * take constraints out the store while necessary                          *
 *                                                                         *
 *-------------------------------------------------------------------------*/
void
CLPR_back()
{

  if (CLPR_level != 0)
    {
      while (vpara->lp->columns + vpara->eqs > CLPR_level)
	{
	  INC_del_constraint(vpara);
	}
      while (vpara->lp->rows > CLPR_index_cmptr)
	{
	  INC_del_variable(vpara);
	}
    }
  else 
    {
      if (vpara != NULL)
	{
	  INC_del_inclp(vpara);
	  vpara=NULL;
	}
    }

}


/*-------------------------------------------------------------------------*
 * CLPR_INIT                                                               *
 *                                                                         *
 * Initialization of the CLP(R) engine                                     *
 *                                                                         *
 *-------------------------------------------------------------------------*/
void
CLPR_Init()
{
  clpr_too_big = Create_Atom("too many CLP(R) variables");

  eq_atom = Tag_ATM(Find_Atom("="));
  le_atom = Tag_ATM(Find_Atom("=<"));
  coma_symbol = Find_Atom(",");
}




void 
CLPR_New_Constraint(WamWord type)
{
  WamWord word, tag_mask;

  CLPR_back();

  if (CLPR_level == 0) 
    {
      vpara=INC_make_inclp();
    }

  DEREF(type, word, tag_mask);

  switch(word)
    {
    case Tag_INT(0) :
      INC_add_new_constraint(vpara, EQ);
      return ;
    case Tag_INT(1) :
       INC_add_new_constraint(vpara, LE);
    }
}
 


void
CLPR_New_Variable(WamWord index, WamWord var)
{

  if (CLPR_index_cmptr > CLPR_VAR_MAX)
    Pl_Err_System(clpr_too_big);         // too many variables all

  CLPR_var[CLPR_index_cmptr] = var;

  INC_add_new_variable(vpara);

  Unify(Tag_INT(CLPR_index_cmptr), index);

  Bind_OV(&CLPR_index_cmptr, CLPR_index_cmptr+1);

  return;
}



void
CLPR_Add_Const(WamWord flt)
{
  WamWord word, tag_mask;
  
  DEREF(flt, word, tag_mask);

  INC_add_to_coef(vpara, 0, -Obtain_Float(UnTag_FLT(word)));
}



void
CLPR_Add_Monom(WamWord index, WamWord mult)
{
  WamWord word, tag_mask;
  
  DEREF(index, word, tag_mask); index = UnTag_INT(word);
  DEREF(mult, word, tag_mask);

  INC_add_to_coef(vpara, index, Obtain_Float(UnTag_FLT(word)));
}

void
CLPR_Add_Monom_to_Cost(WamWord index, WamWord mult)
{
  WamWord word, tag_mask;
  
  DEREF(index, word, tag_mask); index = UnTag_INT(word);
  DEREF(mult, word, tag_mask);
  
  INC_add_to_cost_coef(vpara, index, Obtain_Float(UnTag_FLT(word)));
}


void  
CLPR_Del_Constraint()
{
  INC_del_constraint(vpara);
}



Bool
CLPR_Test()
{
  switch(INC_test_satisfiable(vpara, NULL))
    {
    case TRUE:
      Bind_OV(&CLPR_level, CLPR_level + 1);
      cmptr=vpara->eqs-1;
      return TRUE;
    case TRIVIAL:
      INC_del_constraint(vpara);
      return TRUE;
    }  
  return FALSE;
}


Bool 
CLPR_Minimization(WamWord inf)
{
  double Value; 
  Bool res = FALSE;

  CLPR_back();

  if ((res = (INC_minimize(vpara, &Value))))
    Get_Float(Value, inf);

  INC_cost_to_zero(vpara);
  return res;
}

/*
Bool
CLPR_Test_constant(WamWord index, WamWord Value)
{
  WamWord word, tag_mask;
  double inf, sup;

  DEREF(index, word, tag_mask); 
  word  = UnTag_INT(word);

  INC_add_to_cost_coef(vpara, word, 1.0);

  if (INC_minimize(vpara, &inf)){

    INC_add_to_cost_coef(vpara, word, -2.0);
    
    if (INC_minimize(vpara, &sup)){

      sup = -sup;

      if ((fabs(sup-inf)) <  vpara->lp->epsd) {
	INC_cost_to_zero(vpara);

	sup = (sup + inf)/2; 

	if (fabs(sup - (inf = rint(sup))) < vpara->lp->epsilon)
	  return Get_Float(inf, Value);
	else 
	  return Get_Float(sup, Value);
      }
    }
  }

  INC_cost_to_zero(vpara);
  return FALSE;
}
*/

void
CLPR_cost_to_zero()
{
   INC_cost_to_zero(vpara);
}

void
CLPR_Try_to_detect_instanciation()
{
  matrec ** eq;
  WamWord word, tag_mask;

  eq  = vpara->eq;

  for(
/*       cmptr=vpara->eqs-1 */
      ; cmptr>=0; cmptr--)
    {
      if (CLPR_var[eq[cmptr]->nr.row] != 0)
	{
	  if (eq[cmptr]->next == NULL)
	    {
	      DEREF(CLPR_var[eq[cmptr]->nr.row], word, tag_mask);
	      Bind_OV(CLPR_var+eq[cmptr]->nr.row, 0);
	      if (tag_mask != TAG_FLT_MASK)
		{
		  Get_Float(0.0, word);
		  cmptr--;
		  break;
		}
	    }
	  else if ((eq[cmptr]->next->next == NULL)) 
	    {
	      if (eq[cmptr]->next->nr.row == 0)
		{
		  DEREF(CLPR_var[eq[cmptr]->nr.row], word, tag_mask);
		  Bind_OV(CLPR_var+eq[cmptr]->nr.row, 0);
		  if (tag_mask != TAG_FLT_MASK)
		  {
		    Get_Float(eq[cmptr]->next->value, word);
		    cmptr--;
		    break;
		  }
		}
	      else if (fabs(eq[cmptr]->next->value + 1.0)  < vpara->lp->epsd) 
		{
		  Unify(CLPR_var[eq[cmptr]->nr.row],CLPR_var[eq[cmptr]->next->nr.row]);
		  Bind_OV(CLPR_var+eq[cmptr]->nr.row, 0);
		  cmptr--;
		  break;
		}
	    }
	}
    }
}


Bool
CLPR_test_instanciated(WamWord index)
{
  WamWord word, tag_mask;

  DEREF(index, word, tag_mask);

  return (CLPR_var[UnTag_INT(word)] == 0);
}


Bool
CLPR_Portray(WamWord Res)
{
  int dis, eq;
  WamWord * head, * tail1, * tail2;
  WamWord word, tag_mask;
  matrec * ptr;


  CLPR_back();

  /*
  if (vpara != NULL)
    INC_print_inclp(vpara);
  */

  dis = 1; eq=0;

  DEREF(Res, word, tag_mask);
  if (tag_mask == TAG_REF_MASK)
    if (vpara == NULL) 
      {
	Bind_UV(UnTag_REF(word),NIL_WORD);
	return TRUE;
      }
    else
      Bind_UV(UnTag_REF(word),(WamWord) H);
  else return FALSE;



  tail1 = H++;

  while(( eq < vpara->eqs) || (dis <= vpara->lp->columns))
    {

      if (( eq < vpara->eqs) && (vpara->eq_nr[eq] < dis))
	/* The constraint is an equation */
	{
	  ptr = vpara->eq[eq];

	  if (CLPR_var[ptr->nr.row] == 0) 
	    { eq++; continue;}

	  *tail1 = Tag_LST(H); tail2 = H++; tail1 = H++;
	  
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  *head = eq_atom;
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  //	  if (CLPR_var[ptr->nr.row] == 0) 
	  //	    *head = Put_Float(1.0);
	  //	  else
	    *head = Mk_Tuple(CLPR_var[ptr->nr.row], Put_Float(1.0));
	  ptr = ptr->next; 
	  eq ++;
	}
      else
	/* The constraint is a disequation */
	{
	  *tail1 = Tag_LST(H); tail2 = H++; tail1 = H++;

	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  *head = le_atom;
	  ptr = vpara->lp->mat[dis]; 
	  dis ++;
	}

      for(; ptr != NULL; ptr = ptr->next)
	{
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  if (CLPR_var[ptr->nr.row] == 0) 
	    *head = Put_Float(ptr->value);
	  else
	    *head = Mk_Tuple(CLPR_var[ptr->nr.row], Put_Float(ptr->value));
	}
      *tail2 = NIL_WORD;
    } 
  
  *tail1 = NIL_WORD;

  return TRUE;
}



WamWord
Mk_Tuple(WamWord a, WamWord b)
{
  WamWord *cur_H = H;
  *(H++) = Functor_Arity(coma_symbol, 2);
  *(H++) = a;
  *(H++) = b;
  return Tag_STC(cur_H);
}



Bool
constrainsts_of_intereset(WamWord List, Bool * consts)
{
  WamWord word, tag_mask, * adr;

  Bool * vars;


  DEREF(List, word, tag_mask);
  if (tag_mask == TAG_REF_MASK)
    Pl_Err_Instantiation();

  if (word == NIL_WORD)
    return FALSE;

  vars = calloc(CLPR_VAR_MAX,  sizeof(Bool));
  
  while(tag_mask == TAG_LST_MASK)
    {
      adr = UnTag_LST(word);
      DEREF(Car(adr), word, tag_mask);
      if (tag_mask != TAG_INT_MASK)
	Pl_Err_Type(type_integer, word);
      
      vars[UnTag_INT(word)] = TRUE;
      DEREF(Cdr(adr), word, tag_mask);
    }


  if (word == NIL_WORD)
    {
      free(vars);
      return TRUE;
    }

  free(vars);
  Pl_Err_Type(type_list, List);

  return FALSE;
}



Bool
Finite(WamWord N)
{  
  WamWord word, tag_mask;
  
  DEREF(N, word, tag_mask);

  return (tag_mask == TAG_FLT_MASK && finite(Obtain_Float(UnTag_FLT(N))));
}





