/*-------------------------------------------------------------------------*
 * GNU Prolog                                                              *
 *                                                                         *
 * Part  : CLP(R) Engine                                                   *
 * File  : clpr_portray.pl                                                 *
 * Descr.: portraying of CLP(R) variables                                  *
 * Author: Remy Haemmerle                                                  *
 *                                                                         *
 * Copyright (C) 2001-2002 INRIA, Remy Haemmerle                           *
 *                                                                         *
 * GNU Prolog is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2, or any later version.       *
 *                                                                         *
 * GNU Prolog is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc.  *
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     *
 *-------------------------------------------------------------------------*/


#include "clpr_portray.h"


/*---------------------------------*
 * Constants                       *
 *---------------------------------*/

extern inclp * vpara;

WamWord eq_symbol;
WamWord le_symbol;
int coma_symbol;


/*---------------------------------*/
/* Global Variables                */
/*---------------------------------*/


WamWord Mk_Tuple(WamWord, WamWord);


void
CLPR_Portray_Init()
{
  eq_symbol = Tag_ATM(Find_Atom("="));
  le_symbol = Tag_ATM(Find_Atom("=<"));
  coma_symbol = Find_Atom(",");
}



Bool
CLPR_Portray(WamWord List, WamWord Res)
{
  int dis, eq;
  WamWord * head, * tail1, * tail2;
  matrec * ptr;

  INC_print_inclp(vpara);
  
  dis = 1; eq=0;
  
  {
    WamWord word, tag_mask;
    
    DEREF(Res, word, tag_mask);
    if (tag_mask == TAG_REF_MASK)
      {
	Bind_UV(UnTag_REF(word),H);
      }
    else return FALSE;
  }

  tail1 = H++;

  while(( eq < vpara->eqs) || (dis <= vpara->lp->columns))
    {
      *tail1 = Tag_LST(H); tail2 = H++; tail1 = H++;

      if (( eq < vpara->eqs) && (vpara->eq_nr[eq] < dis))
	/* The constraint is an equation */
	{
	  ptr = vpara->eq[eq];
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  *head = eq_symbol;
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  *head = Mk_Tuple(CLPR_var[ptr->nr.row], Put_Float(1.0));
	  ptr = ptr->next; 
	  eq ++;
	}
      else
	/* The constraint is a disequation */
	{
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  *head = le_symbol;
	  ptr = vpara->lp->mat[dis]; 
	  dis ++;
	}

      for(; ptr != NULL; ptr = ptr->next)
	{
/* 	fatmat[i + ptr->nr.row] = ptr->value; */
	  *tail2 = Tag_LST(H); head = H++; tail2 = H++;
	  *head = Mk_Tuple(CLPR_var[ptr->nr.row], Put_Float(ptr->value));
	}
/*       i += w; */
      *tail2 = NIL_WORD;
    } 
  
  *tail1 = NIL_WORD;

  return TRUE;
}


	/*
void 
initial_variable(WamWord VarsList, Bool * interest_vars)
{

}

	*/


WamWord
Mk_Tuple(WamWord a, WamWord b)
{
  WamWord *cur_H = H;
  *(H++) = Functor_Arity(coma_symbol, 2);
  *(H++) = a;
  *(H++) = b;
  return Tag_STC(cur_H);
}
