/*
    Incremental functions for LP_Solve (header)
    
    Copyright 2002 INRIA, Remy Haemmerle

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _incremental_h 
#define _incremental_h

#include "lpkit.h"

#define INITIAL_EQ    16   /* initial value of eqs_alloc */

#define TRIVIAL -1

typedef struct 
{
  lprec * lp;         /* base of the prb */
  int eqs;            /* numbre of equations in the prb */
  int eqs_alloc;      /* the allocated memory for equations */
  int * eq_nr;        /* eqs_alloc : index of the equation */
  matrec ** eq;       /* eqs_alloc : equations */
  matrec ** svar;     /* eqs_alloc : save of coef to reinvert the matrix
		         if the corresponding equation is deleted*/
  short satisfiable;  /* TRUE is the problem is satisfiable */
}inclp;

/* 
   In a inclp struture all variables are nonbound.
   The inequations are save in dual form in a classical lprec structur.
   The equations are save in eq field and are treated using Gauss Jordan 
   algorithm.
*/


inclp * INC_make_inclp();
/* Allocate an empty incremental problem */

void  INC_del_inclp(inclp * ilp);
/* Desallocate an incremental problem */

void  INC_add_new_variable(inclp * ilp);
/* Add to an incremental problem a variable */

void  INC_add_new_constraint(inclp * ilp, short type);
/* Add a new disequation to a satisfiable problem */

void  INC_del_constraint(inclp * ilp);
/* Delete the last added constraint to an incremental problem */

void  INC_del_variable(inclp * ilp);
/* Delete the last added variable to an incremental problem */

void  INC_add_to_coef(inclp * ilp, int var, REAL Value);
/* Add Value to the coef of the variable Var of the current newconstraint */ 

short INC_test_satisfiable(inclp * ilp, int * svar);
/* Test if the current problem is satisfiable */

void  INC_cost_to_zero(inclp * ilp);
/* Put all coef of the cost function to zero */

void  INC_add_to_cost_coef(inclp * ilp, int var, REAL Value);
/* Add Value to the coef of the variable Var of the cost function */

short INC_minimize(inclp * ilp, REAL * Value);
/* Try to minimize the problem */


/* Incremental Tool functions */

void INC_inc_eq_size(inclp * ilp);

short INC_gauss_pivot(inclp * ilp, int svar);
/* simplify the problem by eliminating a variable */ 

void INC_scale_eq(inclp * lp);
/* scale the new equation by pivot value */

void INC_print_inclp(inclp * lp);
/* print an incremental prb */

/* General Tool functions */

void add_empty_constraint(lprec * lp);
void add_empty_column(lprec *lp);

REAL add_mat(lprec * lp, int Row, int Column, REAL Value);

REAL get_rh(lprec *lp, int Row);
REAL add_rh(lprec * lp, int Row, REAL Value);

matrec * extract_row(lprec * lp, int Row);

#endif
