#include "lpkit.h"
#include "lpglob.h"
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

extern struct inclp * vpara;

#define HASHSIZE 10007

/* Globals */
int     Rows;
int     Columns;
int     Sum;
int     Non_zeros;
int     Level;

REAL	Trej;

short   Maximise;
REAL    Extrad;

int     Warn_count; /* used in CHECK version of rounding macro */

void error(char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  vfprintf(stderr, format, ap);
  fputc('\n', stderr);
  va_end(ap);

  exit(EXIT_FAILURE);
}

lprec *make_lp(int rows, int columns)
{
  lprec *newlp;
  int i, sum;  

#ifndef NO_CHECKING
  if(rows < 0 || columns < 0)
    error("rows < 0 or columns < 0");
#endif

  sum = rows + columns;

  CALLOC(newlp, 1);

  strcpy(newlp->lp_name, "unnamed");

  newlp->verbose = FALSE;
  newlp->print_duals = FALSE;
  newlp->print_sol = FALSE;
  newlp->debug = FALSE;
  newlp->print_at_invert = FALSE;
  newlp->trace = FALSE;

  newlp->rows = rows;
  newlp->columns = columns;
  newlp->sum = sum;
  newlp->sum2 = sum;
  newlp->rows_alloc = rows;
  newlp->columns_alloc = columns;
  newlp->sum_alloc = sum;
  newlp->names_used = FALSE;

  newlp->obj_bound = DEF_INFINITE;
  newlp->infinite = DEF_INFINITE;
  newlp->epsilon = DEF_EPSILON;
  newlp->epsb = DEF_EPSB;
  newlp->epsd = DEF_EPSD;
  newlp->epsel = DEF_EPSEL;
/*    newlp->non_zeros = 0; */
  CALLOC(newlp->col_no, newlp->rows_alloc + 1); 
  CALLOC(newlp->col_no_alloc, newlp->rows_alloc + 1); 
  CALLOC(newlp->row_end, newlp->rows_alloc + 1);

/*    CALLOC(newlp->col_end, newlp->columns_alloc + 1); */
  CALLOC(newlp->mat, newlp->columns_alloc + 1);
  newlp->row_end_valid = TRUE;
  CALLOC(newlp->orig_rh, newlp->rows_alloc + 1);
  CALLOC(newlp->rh, newlp->rows_alloc + 1);
  CALLOC(newlp->rhs, newlp->rows_alloc + 1);

  CALLOC(newlp->must_be_int, newlp->sum_alloc + 1);
  for(i = 0; i <= sum; i++)
    newlp->must_be_int[i] = FALSE;

  CALLOC(newlp->orig_upbo, newlp->sum_alloc + 1);
  for(i = 0; i <= sum; i++)
    newlp->orig_upbo[i] = newlp->infinite;

  CALLOC(newlp->upbo, newlp->sum_alloc + 1);
  CALLOC(newlp->orig_lowbo, newlp->sum_alloc + 1);
  CALLOC(newlp->lowbo, newlp->sum_alloc + 1);

  newlp->basis_valid = TRUE;
  CALLOC(newlp->bas, newlp->rows_alloc + 1);
  CALLOC(newlp->basis, newlp->sum_alloc + 1);
  CALLOC(newlp->lower, newlp->sum_alloc + 1);

  for(i = 0; i <= rows; i++) {
    newlp->bas[i] = i;
    newlp->basis[i] = TRUE;
  }

  for(i = rows + 1; i <= sum; i++)
    newlp->basis[i] = FALSE;


  for(i = 0 ; i <= sum; i++)
    newlp->lower[i] = TRUE;
 
  newlp->eta_valid = TRUE;
  newlp->eta_size = 0;
  newlp->eta_alloc = INITIAL_COL_SIZE*INITIAL_ROW_SIZE;
  newlp->max_num_inv = DEFNUMINV;

  newlp->nr_lagrange = 0;

  CALLOC(newlp->eta_value, newlp->eta_alloc);
  CALLOC(newlp->eta_row_nr, newlp->eta_alloc);

  /* +1 reported by Christian Rank */
  CALLOC(newlp->eta_col_end, newlp->rows_alloc + newlp->max_num_inv + 1);

  newlp->bb_rule = FIRST_NI;
  newlp->break_at_int = FALSE;
  newlp->break_value = 0;

  newlp->iter = 0;
  newlp->total_iter = 0;

  CALLOC(newlp->solution, newlp->sum_alloc + 1);
  CALLOC(newlp->best_solution, newlp->sum_alloc + 1);
  CALLOC(newlp->duals, newlp->rows_alloc + 1);

  newlp->maximise = FALSE;
  newlp->floor_first = TRUE;

  newlp->scaling_used = FALSE;
  newlp->columns_scaled = FALSE;

  CALLOC(newlp->ch_sign, newlp->rows_alloc + 1);

  for(i = 0; i <= rows; i++)
    newlp->ch_sign[i] = FALSE;

  newlp->valid = FALSE; 

  /* create two hash tables for names */
  newlp->rowname_hashtab = create_hash_table(HASHSIZE);
  newlp->colname_hashtab = create_hash_table(HASHSIZE);

  return(newlp);
}

void delete_lp(lprec *lp)
{
  int i; 

  if (lp == NULL) return;

  if(lp->names_used) {
    free(lp->row_name);
    free(lp->col_name);
  }

  for(i = 0; i <= lp->columns; i++)
    free_mat_rec(lp->mat[i]);

  free(lp->mat);

  for(i = 0; i <= lp->rows; i++)
    free(lp->col_no[i]); 
  free(lp->col_no);
  free(lp->col_no_alloc);
 /*free(lp->col_end);*/
  free(lp->row_end);
  free(lp->orig_rh);
  free(lp->rh);
  free(lp->rhs);
  free(lp->must_be_int);
  free(lp->orig_upbo);
  free(lp->orig_lowbo);
  free(lp->upbo);
  free(lp->lowbo);
  free(lp->bas);
  free(lp->basis);
  free(lp->lower);
  free(lp->eta_value);
  free(lp->eta_row_nr);
  free(lp->eta_col_end);
  free(lp->solution);
  free(lp->best_solution);
  free(lp->duals);
  free(lp->ch_sign);
  if(lp->scaling_used)
    free(lp->scale);
  if(lp->nr_lagrange > 0) {
    free(lp->lag_rhs);
    free(lp->lambda);
    free(lp->lag_con_type);
    for(i = 0; i < lp->nr_lagrange; i++)
      free(lp->lag_row[i]);
    free(lp->lag_row);
  }

  /*free_hash_table(lp->rowname_hashtab); 
    free_hash_table(lp->colname_hashtab);*/

  free(lp);

}  

lprec *copy_lp(lprec *lp)
{
  lprec *newlp;
  int i, rowsplus, colsplus, sumplus;
  
  rowsplus = lp->rows_alloc + 1;
  colsplus = lp->columns_alloc + 1;
  sumplus  = lp->sum_alloc + 1;

  MALLOCCPY(newlp, lp, 1); /* copy all non pointers */

  if(newlp->names_used) {
    MALLOCCPY(newlp->col_name, lp->col_name, colsplus);
    MALLOCCPY(newlp->row_name, lp->row_name, rowsplus);
  }

  newlp->rowname_hashtab = copy_hash_table(lp->rowname_hashtab);
  newlp->colname_hashtab = copy_hash_table(lp->colname_hashtab);

  MALLOCCPY(newlp->mat_alloc, lp->mat_alloc, colsplus)
  MALLOC(newlp->mat, colsplus);
  for(i = 0; i < colsplus; i++)
    newlp->mat[i] = copy_mat_rec(lp->mat[i]);
/*    MALLOCCPY(newlp->col_end, lp->col_end, colsplus); */
  MALLOCCPY(newlp->col_no_alloc, lp->col_no_alloc, rowsplus);
  MALLOC(newlp->col_no, rowsplus);
  for(i = 0; i < rowsplus; i++) 
    MALLOCCPY(newlp->col_no[i], lp->col_no[i], lp->col_no_alloc[i]); 
  MALLOCCPY(newlp->row_end, lp->row_end, rowsplus);
  MALLOCCPY(newlp->orig_rh, lp->orig_rh, rowsplus);
  MALLOCCPY(newlp->rh, lp->rh, rowsplus);
  MALLOCCPY(newlp->rhs, lp->rhs, rowsplus);
  MALLOCCPY(newlp->must_be_int, lp->must_be_int, sumplus);
  MALLOCCPY(newlp->orig_upbo, lp->orig_upbo, sumplus);
  MALLOCCPY(newlp->orig_lowbo, lp->orig_lowbo, sumplus);
  MALLOCCPY(newlp->upbo, lp->upbo, sumplus);
  MALLOCCPY(newlp->lowbo, lp->lowbo, sumplus);
  MALLOCCPY(newlp->bas, lp->bas, rowsplus);
  MALLOCCPY(newlp->basis, lp->basis, sumplus);
  MALLOCCPY(newlp->lower, lp->lower, sumplus);
  MALLOCCPY(newlp->eta_value, lp->eta_value, lp->eta_alloc);
  MALLOCCPY(newlp->eta_row_nr, lp->eta_row_nr, lp->eta_alloc);
  MALLOCCPY(newlp->eta_col_end, lp->eta_col_end,
	    lp->rows_alloc + lp->max_num_inv + 1);
  MALLOCCPY(newlp->solution, lp->solution, sumplus);
  MALLOCCPY(newlp->best_solution, lp->best_solution, sumplus);
  MALLOCCPY(newlp->duals, lp->duals, rowsplus);
  MALLOCCPY(newlp->ch_sign, lp->ch_sign, rowsplus);

  if(newlp->scaling_used)
    MALLOCCPY(newlp->scale, lp->scale, sumplus);

  if(newlp->nr_lagrange > 0) {
    MALLOCCPY(newlp->lag_rhs, lp->lag_rhs, newlp->nr_lagrange);
    MALLOCCPY(newlp->lambda, lp->lambda, newlp->nr_lagrange);
    MALLOCCPY(newlp->lag_con_type, lp->lag_con_type, newlp->nr_lagrange);
    MALLOC(newlp->lag_row, newlp->nr_lagrange);
    for(i = 0; i < newlp->nr_lagrange; i++)
      MALLOCCPY(newlp->lag_row[i], lp->lag_row[i], colsplus);
  }
  return(newlp);
}

void inc_col_no_space(lprec * lp, int Row)
{

#ifndef NO_CHECKING
  if(Row > lp->rows || Row < 0)
    error("Row out of range");
#endif

  if(lp->row_end[Row] >= lp->col_no_alloc[Row])
    {
      if(lp->col_no_alloc[Row] < INITIAL_ROW_SIZE)
	{
	  lp->col_no_alloc[Row] = INITIAL_ROW_SIZE;
	}
      else
	{
	  lp->col_no_alloc[Row] *= 1.5;
	}
      
      REALLOC(lp->col_no[Row], lp->col_no_alloc[Row]);
    }

}
 void inc_row_space(lprec *lp)
{
  int i, diff;

  int rows = lp->rows_alloc;
  int sum  = lp->sum_alloc;

  if(lp->rows >= lp->rows_alloc) {
    if (lp->rows_alloc < INITIAL_ROW_SIZE) 
      {
	lp->rows_alloc = INITIAL_ROW_SIZE;
      }
    else
      {
	lp->rows_alloc = lp->rows*1.5;
      }

    diff = lp->rows_alloc - rows;

    lp->sum_alloc  = lp->rows_alloc + lp->columns_alloc;
    REALLOC(lp->orig_rh, lp->rows_alloc + 1);
    REALLOC(lp->rh, lp->rows_alloc + 1);
    REALLOC(lp->rhs, lp->rows_alloc + 1);
    REALLOC(lp->row_end, lp->rows_alloc + 1);
    REALLOC(lp->bas, lp->rows_alloc + 1);
    REALLOC(lp->duals, lp->rows_alloc + 1);
    REALLOC(lp->ch_sign, lp->rows_alloc + 1);
    REALLOC(lp->col_no, lp->rows_alloc + 1);
    REALLOC(lp->col_no_alloc, lp->rows_alloc + 1);
    REALLOC(lp->eta_col_end, lp->rows_alloc + lp->max_num_inv + 1);
    if(lp->names_used)
      REALLOC(lp->row_name, lp->rows_alloc + 1);
    if(lp->scaling_used)
        REALLOC_SUM(lp->scale, sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->orig_upbo    , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->upbo         , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->orig_lowbo   , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->lowbo        , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->solution     , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->best_solution, sum, lp->sum_alloc+1, rows, lp->rows_alloc);
 /*     REALLOC_SUM(lp->masked       , sum, lp->sum_alloc+1, rows, lp->rows_alloc); */
    REALLOC_SUM(lp->basis        , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->lower        , sum, lp->sum_alloc+1, rows, lp->rows_alloc);
    REALLOC_SUM(lp->must_be_int  , sum, lp->sum_alloc+1, rows, lp->rows_alloc);

    lp->sum2 = lp->rows_alloc+lp->columns;

    for(i=0; i<=lp->rows; i++)
      if (lp->bas[i] > rows)
	lp->bas[i] += diff;

 /*    for(i=lp->rows_alloc; i<lp->sum2; i++) */
/*        if (lp->bas[i] > rows) */
/*  	lp->bas[i] += diff; */
  }

}

void inc_col_space(lprec *lp)
{
  if(lp->columns >= lp->columns_alloc) {
    if (lp->columns_alloc < INITIAL_COL_SIZE) 
      lp->columns_alloc = INITIAL_COL_SIZE;
    else
      lp->columns_alloc = lp->columns_alloc*1.5;
    lp->sum_alloc = lp->rows_alloc + lp->columns_alloc;

    REALLOC(lp->must_be_int, lp->sum_alloc + 1);
    REALLOC(lp->orig_upbo, lp->sum_alloc + 1);
    REALLOC(lp->upbo, lp->sum_alloc + 1);
    REALLOC(lp->orig_lowbo, lp->sum_alloc + 1);
    REALLOC(lp->lowbo, lp->sum_alloc + 1);
    REALLOC(lp->solution, lp->sum_alloc + 1);
    REALLOC(lp->best_solution, lp->sum_alloc + 1);
 /*     REALLOC(lp->masked, lp->sum_alloc + 1); */
    REALLOC(lp->basis, lp->sum_alloc + 1);
    REALLOC(lp->lower, lp->sum_alloc + 1);
    if(lp->names_used)
      REALLOC(lp->col_name, lp->columns_alloc + 1);
    if(lp->scaling_used)
      REALLOC(lp->scale, lp->sum_alloc + 1);
/*      REALLOC(lp->col_end, lp->columns_alloc + 1); */
    REALLOC(lp->mat, lp->columns_alloc + 1);
    REALLOC(lp->mat_alloc, lp->columns_alloc + 1)

    lp->sum2 = lp->rows_alloc+lp->columns;
  }

}

void set_mat(lprec *lp, int Row, int Column, REAL Value)
{
  matrec ** ptr; int i, j;

  /*stop(25);*/
  /* This function is very inefficient if used to add new matrix entries in
     other places than at the end of the matrix. OK for replacing existing
     non-zero values */

#ifndef NO_CHECKING
  if(Row > lp->rows || Row < 0)
    error("Row out of range");
  if(Column > lp->columns || Column < 1)
    error("Column out of range");
#endif   

  /* scaling is performed twice? MB */
  if(lp->scaling_used)
    Value *= lp->scale[Row] * lp->scale[lp->rows_alloc + Column];
  
  if (lp->basis[Column+lp->rows_alloc] == TRUE && Row > 0)
    lp->basis_valid = FALSE;
  lp->eta_valid = FALSE;

  /* find out if we already have such an entry */
  FIND_MAT(ptr, lp->mat[Column], Row);

  if( ((*ptr) != NULL) && ((*ptr)->nr.row == Row)) {
    /* there is an existing entry */
    if(my_abs(Value) > lp->epsilon) { /* we replace it by something non-zero */
      if (lp->scaling_used) {
	if(lp->ch_sign[Row])
	  (*ptr)->value = -Value * lp->scale[Row] * lp->scale[Column];
	else
	  (*ptr)->value = Value * lp->scale[Row] * lp->scale[Column];
      }
      else { /* no scaling */
	if(lp->ch_sign[Row])
	  (*ptr)->value = -Value;
	else
	  (*ptr)->value = Value;
      }
    }
    else { /* setting existing non-zero entry to zero. Remove the entry */
      /* this might remove an entire column, or leave just a bound. No
	 nice solution for that yet */
      FREE_MAT(ptr);
       
      if (lp->row_end_valid)
	{
	  lp->row_end[Row]--;
	  i = lp->row_end[Row];
	  while((lp->col_no[Row][i] != Column)) i--;
	  
	  for(j = i; j < lp->row_end[Row] ; j++)
	    lp->col_no[Row][j] =  lp->col_no[Row][j+1];
	}
    }
  }
  else if(my_abs(Value) > lp->epsilon) {
    /* no existing entry. make new one only if not nearly zero */
    /* check if more space is needed for matrix */

    if (lp->scaling_used) {
      if(lp->ch_sign[Row])
	INSERT_MAT(*ptr ,Row, -Value * lp->scale[Row] * lp->scale[Column])
      else
	INSERT_MAT(*ptr, Row, Value * lp->scale[Row] * lp->scale[Column])
    }
    else /* no scaling */
      {
	if(lp->ch_sign[Row])
	  INSERT_MAT(*ptr, Row, -Value)
	else
	  INSERT_MAT(*ptr, Row, Value)
      }
    
    if (lp->row_end_valid)
      {
	i = lp->row_end[Row]-1;
	lp->row_end[Row]++;
	while((i>=0) && (lp->col_no[Row][i]>Column)) i--;
	 
	inc_col_no_space(lp, Row);
	for(j = lp->row_end[Row]-2; j > i; j--)
	  lp->col_no[Row][j+1] =  lp->col_no[Row][j];


	lp->col_no[Row][i+1] = Column;
      }
/*      lp->non_zeros++; */
  }      
}

void set_obj_fn(lprec *lp, REAL *row)
{
  int i, rows;
  rows = lp->rows_alloc;
  for(i = 1; i <= lp->columns; i++)
    set_mat(lp, 0, i, row[i]);
}       

void str_set_obj_fn(lprec *lp, char *row)
{
  int  i;
  REAL *arow;
  char *p, *newp;
  CALLOC(arow, lp->columns + 1);
  p = row;
  for(i = 1; i <= lp->columns; i++) {
    arow[i] = (REAL) strtod(p, &newp);
    if(p == newp)
      error("Bad string in str_set_obj_fn");
    else
      p = newp; 
  }
  set_obj_fn(lp, arow);
  free(arow);
}


void add_constraint(lprec *lp, REAL *row, short constr_type, REAL rh)
{
  int  i;
  matrec ** ptr;

  lp->rows++;
  lp->sum++;
  inc_row_space(lp);

  lp->row_end[lp->rows] = 0;
  lp->col_no_alloc[lp->rows] = INITIAL_ROW_SIZE;
  MALLOC(lp->col_no[lp->rows], INITIAL_ROW_SIZE);

/*    if(lp->scaling_used) { */
      /* shift scale */ 
/*      for(i = lp->sum; i > lp->rows_alloc; i--) */
/*        lp->scale[i] = lp->scale[i - 1]; */
/*      lp->scale[lp->rows] = 1; */
/*    } */

  if(lp->names_used)
    sprintf(lp->row_name[lp->rows], "r_%d", lp->rows);

    
  if(constr_type == GE)
    lp->ch_sign[lp->rows] = TRUE;
  else
    lp->ch_sign[lp->rows] = FALSE;

    for(i = 1; i <= lp->columns; i++)
      if(row[i] != 0) 
	{
	  FIND_MAT(ptr, lp->mat[i], lp->rows); 
	  INSERT_MAT(*ptr, lp->rows, row[i]) 
	  if(lp->scaling_used && lp->columns_scaled)
	    (*ptr)->value *= lp->scale[lp->rows_alloc + i];
	  if(lp->ch_sign[lp->rows])
	    (*ptr)->value *= -1; 
/*  	  lp->non_zeros++; */
/*  	  lp->col_end[i]++; */
	}

/*    for(i = lp->sum; i > lp->rows_alloc; i--) { */
/*      lp->orig_upbo[i]   = lp->orig_upbo[i - 1]; */
/*      lp->orig_lowbo[i]  = lp->orig_lowbo[i - 1]; */
/*      lp->masked[i]      = lp->masked[i - 1]; */
/*      lp->basis[i]       = lp->basis[i - 1]; */
/*      lp->lower[i]       = lp->lower[i - 1]; */
/*      lp->must_be_int[i] = lp->must_be_int[i - 1]; */
/*    } */

  /* changed from i <= lp->rows to i < lp->rows, MB */
/*    for(i = 1 ; i < lp->rows; i++) */
/*      if(lp->bas[i] >= lp->rows) */
/*        lp->bas[i]++; */

  if(constr_type == LE || constr_type == GE) {
    lp->orig_upbo[lp->rows] = lp->infinite;
  }
  else if(constr_type == EQ) {
    lp->orig_upbo[lp->rows] = 0;
  }
  else {
    fprintf(stderr, "Wrong constraint type\n");
    exit(EXIT_FAILURE);
  }

  lp->orig_lowbo[lp->rows] = 0;

  if(constr_type == GE && rh != 0)
    lp->orig_rh[lp->rows] = -rh;
  else
    lp->orig_rh[lp->rows] = rh;  

  lp->row_end_valid = FALSE;
 
  lp->bas[lp->rows] = lp->rows;
  lp->basis[lp->rows] = TRUE;
  lp->lower[lp->rows] = TRUE;   
  lp->eta_valid = FALSE;
}

void str_add_constraint(lprec *lp,
			char *row_string,
			short constr_type,
			REAL rh)
{
  int  i;
  REAL *aRow;
  char *p, *newp;
  CALLOC(aRow, lp->columns + 1);
  p = row_string;
 
  for(i = 1; i <= lp->columns; i++) {
    aRow[i] = (REAL) strtod(p, &newp);
    if(p == newp)
      error("Bad string in str_add_constr");
    else
      p = newp; 
  }
  add_constraint(lp, aRow, constr_type, rh);
  free(aRow);
}

void del_constraint(lprec *lp, int del_row)
{
  int i;
  matrec ** ptr;

#ifndef NO_CHECKING
  if(del_row<1 || del_row>lp->rows)
    {
      fprintf(stderr, "There is no constraint nr. %d\n", del_row);
      exit(EXIT_FAILURE);
    }
#endif

  for(i = 1; i <= lp->columns; i++) 
    {
      FIND_MAT(ptr, lp->mat[i], del_row);

      if( ((*ptr) != NULL) && ((*ptr)->nr.row == del_row)) 
	FREE_MAT(ptr);

      for(; (*ptr) != NULL; ptr = &((*ptr)->next))
	(*ptr)->nr.row -= 1;
      
    }

  for(i = del_row; i < lp->rows; i++) {
    lp->orig_rh[i] = lp->orig_rh[i + 1];
    lp->ch_sign[i] = lp->ch_sign[i + 1];
    lp->bas[i] = lp->bas[i + 1];
    if(lp->names_used)
      strcpy(lp->row_name[i], lp->row_name[i + 1]);
  }
  for(i = 1; i < lp->rows; i++)
    if((lp->bas[i] >  del_row) && (lp->bas[i] < lp->rows))
      lp->bas[i]--;

  for(i = del_row; i < lp->rows_alloc; i++) {
    lp->lower[i] = lp->lower[i + 1];
    lp->basis[i] = lp->basis[i + 1];
/*      lp->masked[i] = lp->masked[i + 1]; */
    lp->orig_upbo[i] = lp->orig_upbo[i + 1];
    lp->orig_lowbo[i] = lp->orig_lowbo[i + 1];
    lp->must_be_int[i] = lp->must_be_int[i + 1];
    if(lp->scaling_used)
      lp->scale[i] = lp->scale[i + 1];
  }

  free(lp->col_no[lp->rows]);
  lp->col_no_alloc[lp->rows] = 0;

  lp->rows--;
  lp->sum--; 

  /*lp->row_end_valid = FALSE;*/
  lp->eta_valid     = FALSE;
  lp->basis_valid   = FALSE; 
}

void add_lag_con(lprec *lp, REAL *row, short con_type, REAL rhs)
{
  int i;
  REAL sign;
  if(con_type == LE || con_type == EQ)
    sign = 1;
  else if(con_type == GE)
    sign = -1;
  else
    error("con_type not implemented\n");

  lp->nr_lagrange++;
  if(lp->nr_lagrange == 1) {
    CALLOC(lp->lag_row, lp->nr_lagrange);
    CALLOC(lp->lag_rhs, lp->nr_lagrange);
    CALLOC(lp->lambda, lp->nr_lagrange);
    CALLOC(lp->lag_con_type, lp->nr_lagrange);
  }
  else {
    REALLOC(lp->lag_row, lp->nr_lagrange);
    REALLOC(lp->lag_rhs, lp->nr_lagrange);
    REALLOC(lp->lambda, lp->nr_lagrange);
    REALLOC(lp->lag_con_type, lp->nr_lagrange);
  }
  CALLOC(lp->lag_row[lp->nr_lagrange-1], lp->columns+1);
  lp->lag_rhs[lp->nr_lagrange-1] = rhs * sign;
  for( i = 1; i <= lp->columns; i++)
    lp->lag_row[lp->nr_lagrange-1][i] = row[i] * sign;
  lp->lambda[lp->nr_lagrange-1] = 0;
  lp->lag_con_type[lp->nr_lagrange-1]=(con_type == EQ);
}

void str_add_lag_con(lprec *lp, char *row, short con_type, REAL rhs)
{
  int  i;
  REAL *a_row;
  char *p, *new_p;
  CALLOC(a_row, lp->columns + 1);
  p = row;
 
  for(i = 1; i <= lp->columns; i++) {
    a_row[i] = (REAL) strtod(p, &new_p);
    if(p == new_p)
      error("Bad string in str_add_lag_con");
    else
      p = new_p; 
  }
  add_lag_con(lp, a_row, con_type, rhs);
  free(a_row);
}


void add_column(lprec *lp, REAL *column)
{
  int i;

  matrec ** ptr; 

  /* if the column has only one entry, this should be handled as a bound, but
     this currently is not the case */

  lp->columns++;
  lp->sum++;
  lp->sum2++;
  inc_col_space(lp);
  lp->mat[lp->columns]=NULL;

  ptr = &(lp->mat[lp->columns]);

/*    lp->col_end[lp->columns] = 0; */
/*    lp->mat_alloc[lp->columns] = 0; */

  if(lp->scaling_used)
    {
      for(i = 0; i <= lp->rows; i++)
	column[i] *= lp->scale[i];
      lp->scale[lp->sum2] = 1;
      
      for(i = 0 ; i <= lp->rows ; i++)
	if(column[i] != 0) {
	  if(lp->ch_sign[i])
	    INSERT_MAT((*ptr), i, -column[i] * lp->scale[i])
	  else
	    INSERT_MAT((*ptr), i, column[i] * lp->scale[i])
	  ptr = &((*ptr)->next);
	}
    }
  else
    {
      for(i = 0 ; i <= lp->rows ; i++)
	if(column[i] != 0) {
	  if(lp->ch_sign[i])
	    INSERT_MAT((*ptr), i, -column[i])
	  else
	    INSERT_MAT((*ptr), i, column[i])
	  ptr = &((*ptr)->next);
	}
    }

  /*   lp->col_end[lp->columns] = elmnr; */
  /*    lp->non_zeros += elmnr; */
  lp->orig_lowbo[lp->sum2] = 0;
  lp->orig_upbo[lp->sum2] = lp->infinite;
  lp->lower[lp->sum2] = TRUE;
  lp->basis[lp->sum2] = FALSE;
/*    lp->masked[lp->sum2]= FALSE; */
  lp->must_be_int[lp->sum2] = FALSE;
  if(lp->names_used)
    sprintf(lp->col_name[lp->columns], "var_%d", lp->columns);
 
  lp->row_end_valid = FALSE;
}

void str_add_column(lprec *lp, char *col_string)
{
  int  i;
  REAL *aCol;
  char *p, *newp;
  CALLOC(aCol, lp->rows + 1);
  p = col_string;
 
  for(i = 0; i <= lp->rows; i++) {
    aCol[i] = (REAL) strtod(p, &newp);
    if(p == newp)
      error("Bad string in str_add_column");
    else
      p = newp; 
  }
  add_column(lp, aCol);
  free(aCol);
}

void del_column(lprec *lp, int column)
{
  int i, j;

  if(column > lp->columns || column < 1)
    error("Column out of range in del_column");
  for(i = 1; i <= lp->rows; i++) {
    if(lp->bas[i] == lp->rows_alloc + column)
      lp->basis_valid = FALSE;
    else if(lp->bas[i] > lp->rows_alloc + column)
      lp->bas[i]--;
  }
  for(i = lp->rows_alloc + column; i < lp->sum2; i++) {
    if(lp->names_used)
      strcpy(lp->col_name[i - lp->rows_alloc], lp->col_name[i - lp->rows_alloc + 1]);
    lp->must_be_int[i] = lp->must_be_int[i + 1];
    lp->orig_upbo[i] = lp->orig_upbo[i + 1];
    lp->orig_lowbo[i] = lp->orig_lowbo[i + 1];
    lp->upbo[i] = lp->upbo[i + 1];
    lp->lowbo[i] = lp->lowbo[i + 1];
/*      lp->masked[i]= lp->masked[i + 1]; */
    lp->basis[i] = lp->basis[i + 1];
    lp->lower[i] = lp->lower[i + 1];
    if(lp->scaling_used)
      lp->scale[i] = lp->scale[i + 1];
  }
  for(i = 0; i < lp->nr_lagrange; i++)
    for(j = column; j <= lp->columns; j++)
      lp->lag_row[i][j] = lp->lag_row[i][j+1];

 /*   lp->non_zeros -= lp->col_end[column]; */

  free_mat_rec(lp->mat[column]);
  for(i = column; i < lp->columns; i++)
  {
/*      lp->col_end[i] = lp->col_end[i + 1]; */
    lp->mat[i]=lp->mat[i+1];
/*      lp->mat_alloc[i] = lp->mat_alloc[i+1]; */
  }

/*    lp->mat_alloc[column] = 0; */

  lp->row_end_valid = FALSE;
  lp->eta_valid = FALSE;

  lp->sum--;
  lp->sum2--;
  lp->columns--;
}


void set_upbo(lprec *lp, int column, REAL value)
{
  if(column > lp->columns || column < 1)
    error("Column out of range");
  if(lp->scaling_used)
    value /= lp->scale[lp->rows_alloc + column];
  if(value < lp->orig_lowbo[lp->rows_alloc + column])
    error("Upperbound must be >= lowerbound"); 
  lp->eta_valid = FALSE;
  lp->orig_upbo[lp->rows_alloc + column] = value;
}

void set_lowbo(lprec *lp, int column, REAL value)
{
  if(column > lp->columns || column < 1)
    error("Column out of range");
  if(lp->scaling_used)
    value /= lp->scale[lp->rows_alloc + column];
  if(value > lp->orig_upbo[lp->rows_alloc + column])
    error("Upperbound must be >= lowerbound"); 
  /*
    if(value < 0)
    error("Lower bound cannot be < 0");
    */
  lp->eta_valid = FALSE;
  lp->orig_lowbo[lp->rows_alloc + column] = value;
}

void set_int(lprec *lp, int column, short must_be_int)
{
  if(column > lp->columns || column < 1)
    error("Column out of range");
  lp->must_be_int[lp->rows_alloc + column] = must_be_int;
  if(must_be_int == TRUE)
    if(lp->columns_scaled)
      unscale_columns(lp);
}


void set_rh(lprec *lp, int row, REAL value)
{
  if(row > lp->rows || row < 0)
    error("Row out of Range");

  if ((row == 0) && (!lp->maximise))  /* setting of RHS of OF IS meaningful */
    value = -value;
  if(lp->scaling_used) {
    if(lp->ch_sign[row])
      lp->orig_rh[row] = -value * lp->scale[row];
    else
      lp->orig_rh[row] = value * lp->scale[row];
  }
  else
    if(lp->ch_sign[row])
      lp->orig_rh[row] = -value;
    else
      lp->orig_rh[row] = value;
  lp->eta_valid = FALSE;
} 

void set_rh_vec(lprec *lp, REAL *rh)
{
  int i;
  if(lp->scaling_used) {
    for(i = 1; i <= lp->rows; i++)
      if(lp->ch_sign[i])
        lp->orig_rh[i] = -rh[i]*lp->scale[i];
      else
        lp->orig_rh[i] = rh[i]*lp->scale[i];
  }
  else
    for(i = 1; i <= lp->rows; i++)
      if(lp->ch_sign[i])
        lp->orig_rh[i] = -rh[i];
      else
        lp->orig_rh[i] = rh[i];
  lp->eta_valid = FALSE;
}

void str_set_rh_vec(lprec *lp, char *rh_string)
{
  int  i;
  REAL *newrh;
  char *p, *newp;
  CALLOC(newrh, lp->rows + 1);
  p = rh_string;
 
  for(i = 1; i <= lp->rows; i++) {
    newrh[i] = (REAL) strtod(p, &newp);
    if(p == newp)
      error("Bad string in str_set_rh_vec");
    else
      p = newp; 
  }
  set_rh_vec(lp, newrh);
  free(newrh);
}


void set_maxim(lprec *lp)
{
  int i;
  if(lp->maximise == FALSE) {
    for(i = 0; i <= lp->columns; i++)
      if(lp->mat[i] != 0)
	if(lp->mat[i]->nr.row == 0)
	  lp->mat[i]->value *= -1;
    lp->eta_valid = FALSE;
    lp->orig_rh[0] *= -1;
  } 
  lp->maximise = TRUE;
  lp->ch_sign[0] = TRUE;
}

void set_minim(lprec *lp)
{
  int i;
  if(lp->maximise == TRUE) {
    for(i = 0; i <= lp->columns; i++)
      if(lp->mat != NULL)
	if(lp->mat[i]->nr.row == 0)
	  lp->mat[i]->value *= -1;
    lp->eta_valid = FALSE;
    lp->orig_rh[0] *= -1;
  } 
  lp->maximise = FALSE;
  lp->ch_sign[0] = FALSE;
}

void set_constr_type(lprec *lp, int row, short con_type)
{
  int i;
  matrec ** ptr;
  

#ifndef NO_CHECKING
  if(row > lp->rows || row < 1)
    error("Row out of Range");
#endif

  if(con_type == EQ) {
    lp->orig_upbo[row] = 0;
    lp->basis_valid = FALSE;
    if(lp->ch_sign[row]) {
      for(i = 0; i <= lp->columns; i++)
	{
	  FIND_MAT(ptr, lp->mat[i], row); 
	  if( ((*ptr) != NULL) && ((*ptr)->nr.row == row)) 
	    (*ptr)->value *= -1;
	}
      lp->eta_valid = FALSE;
      lp->ch_sign[row] = FALSE;
      if(lp->orig_rh[row] != 0)
	lp->orig_rh[row] *= -1;
    }
  }
  else if(con_type == LE) {
    lp->orig_upbo[row] = lp->infinite;
    lp->basis_valid = FALSE;
    if(lp->ch_sign[row]) {
      for(i = 0; i <= lp->columns; i++)
	{
	  FIND_MAT(ptr, lp->mat[i], row); 
	  if( ((*ptr) != NULL) && ((*ptr)->nr.row == row)) 
	    (*ptr)->value *= -1;
	}
      lp->eta_valid = FALSE;
      lp->ch_sign[row] = FALSE;
      if(lp->orig_rh[row] != 0)
	lp->orig_rh[row] *= -1;
    }
  }
  else if(con_type == GE) {
    lp->orig_upbo[row] = lp->infinite;
    lp->basis_valid = FALSE;
    if(!lp->ch_sign[row]) {
      for(i = 0; i <= lp->columns; i++)
	{
	  FIND_MAT(ptr, lp->mat[i], row); 
	  if( ((*ptr) != NULL) && ((*ptr)->nr.row == row)) 
	    (*ptr)->value *= -1;
	}
      lp->eta_valid = FALSE;
      lp->ch_sign[row] = TRUE;
      if(lp->orig_rh[row] != 0)
	lp->orig_rh[row] *= -1;
    }
  } 
  else
    error("Constraint type not (yet) implemented");
}

REAL mat_elm(lprec *lp, int row, int column)
{
  REAL value;
  matrec ** ptr;

#ifndef NO_CHECKING
  if(row < 0 || row > lp->rows)
    error("Row out of range in mat_elm");
  if(column < 1 || column > lp->columns)
    error("Column out of range in mat_elm");
#endif

  value = 0;
/*     j = lp->col_end[column-1]; */
/*      while(lp->mat[elmnr].row_nr != row && elmnr < lp->col_end[column]) */
/*    elmnr++; */
  FIND_MAT(ptr, lp->mat[column], row); 
  if( ((*ptr) != NULL) && ((*ptr)->nr.row == row)) 
    {
      value = (*ptr)->value;
      if(lp->ch_sign[row])
	value = -value;
      if(lp->scaling_used)
	value /= lp->scale[row] * lp->scale[lp->rows_alloc + column];
    }
  return(value);
}

void get_row(lprec *lp, int row_nr, REAL *row)
{
  int i;
  matrec ** ptr;

#ifndef NO_CHECKING
  if(row_nr <0 || row_nr > lp->rows)
    error("Row nr. out of range in get_row");
#endif

  row[0]=0;
  for(i = 1; i <= lp->columns; i++)
    {
      FIND_MAT(ptr, lp->mat[i], row_nr); 
      if( ((*ptr) != NULL) && ((*ptr)->nr.row == row_nr)) 
	{
	  row[i] = (*ptr)->value;
	  if(lp->scaling_used)
	    row[i] /= lp->scale[lp->rows_alloc + i] * lp->scale[row_nr];
	  if(lp->ch_sign[row_nr])
	    row[i] = -row[i];
	}
      else
	row[i]=0;
    }
}

void get_column(lprec *lp, int col_nr, REAL *column)
{
  int i;
  matrec * ptr;

#ifndef NO_CHECKING
  if(col_nr < 1 || col_nr > lp->columns)
    error("Col. nr. out of range in get_column");
#endif

  for(i = 0; i <= lp->rows; i++)
    column[i] = 0;
 
  for(ptr = lp->mat[col_nr]; ptr != NULL; ptr = ptr->next)
    column[ptr->nr.row] = ptr->value;

  for(i = 0; i <= lp->rows; i++)
    if(column[i] != 0) {
      if(lp->ch_sign[i])
	column[i] *= -1;
      if(lp->scaling_used)
	column[i] /= (lp->scale[i] * lp->scale[lp->rows_alloc + col_nr]);
    }
}


void get_reduced_costs(lprec *lp, REAL *rc)
{
  int varnr, i;
  REAL f;
  matrec * ptr;

  if(!lp->basis_valid)
    error("Not a valid basis in get_reduced_costs");

  if(!lp->eta_valid)
    invert(lp);  

  for(i = 1; i <= lp->sum; i++) 
    rc[i] = 0;
  rc[0] = 1;

  btran(lp, rc);

  for(i = 1; i <= lp->columns; i++) {
    varnr = lp->rows_alloc + i;
    if(!lp->basis[varnr])
      if(lp->upbo[varnr] > 0) {
	f = 0;
	for(ptr = lp->mat[i]; ptr != NULL; ptr = ptr->next)
	  f += rc[ptr->nr.row] * ptr->value;

	rc[lp->rows + i] = f;
      }
  }
  for(i = 1; i <= lp->sum; i++) 
    my_round(rc[i], lp->epsd);
}   

short is_feasible(lprec *lp, REAL *values)
{
  int i;
  REAL *this_rhs;
  REAL dist;
  matrec * ptr;

  if(lp->scaling_used) {
    for(i = lp->rows_alloc + 1; i <= lp->sum2; i++)
      if(   values[i - lp->rows_alloc] < lp->orig_lowbo[i] * lp->scale[i]
	 || values[i - lp->rows_alloc] > lp->orig_upbo[i]  * lp->scale[i])
	return(FALSE);
  }
  else {
    for(i = lp->rows_alloc + 1; i <= lp->sum2; i++)
      if(   values[i - lp->rows_alloc] < lp->orig_lowbo[i]
	 || values[i - lp->rows_alloc] > lp->orig_upbo[i])
	return(FALSE);
  }
  CALLOC(this_rhs, lp->rows + 1);
  if (lp->columns_scaled) {
    for(i = 1; i <= lp->columns; i++) 
      for(ptr = lp->mat[i]; ptr != NULL; ptr = ptr->next)
	   this_rhs[ptr->nr.row] += 
	     ptr->value * values[i] / lp->scale[lp->rows_alloc + i];
  }
  else {
    for(i = 1; i <= lp->columns; i++)
      for(ptr = lp->mat[i]; ptr != NULL; ptr = ptr->next)
	this_rhs[ptr->nr.row] += ptr->value * values[i];
  }
  for(i = 1; i <= lp->rows; i++) {
    dist = lp->orig_rh[i] - this_rhs[i];
    my_round(dist, 0.001); /* ugly constant, MB */
    if((lp->orig_upbo[i] == 0 && dist != 0) || dist < 0) {
      free(this_rhs);
      return(FALSE);
    }     
  } 
  free(this_rhs);
  return(TRUE);
}

/* fixed by Enrico Faggiolo */
short column_in_lp(lprec *lp, REAL *testcolumn)
{
  int i;
  int nz, ident;
  REAL value;
  matrec * ptr;

  for(nz = 0, i = 0; i <= lp->rows; i++)
    if(my_abs(testcolumn[i]) > lp->epsel) nz++;  

  if(lp->scaling_used)
    for(i = 1; i <= lp->columns; i++) {
      ident = nz;
      for(ptr = lp->mat[i]; ptr != NULL; ptr = ptr->next)
	{
	  value = ptr->value;
	  if(lp->ch_sign[ptr->nr.row])
	    value = -value;
	  value /= lp->scale[lp->rows_alloc + i];
	  value /= lp->scale[ptr->nr.row];
	  value -= testcolumn[ptr->nr.row];
	  if(my_abs(value) > lp->epsel)
	    break;
	  ident--;
	  if(ident == 0)
	    return(TRUE);
	}
    }
  else
    for(i = 1; i <= lp->columns; i++) {
      ident = nz;
      for(ptr = lp->mat[i]; ptr != NULL; ptr = ptr->next)
	{
	  value = ptr->value;
	  if(lp->ch_sign[ptr->nr.row])
	  value = -value;
	  value -= testcolumn[ptr->nr.row];
	  if(my_abs(value) > lp->epsel)
	    break;
	  ident--;
	  if(ident == 0)
	    return(TRUE);
	}
    }
  return(FALSE);
}

void print_lp(lprec *lp)
{
  int i, j;
  REAL *fatmat;
  matrec * ptr;

  CALLOC(fatmat, (lp->rows + 1) * lp->columns);
  for(i = 1; i <= lp->columns; i++)
    for(ptr = lp->mat[i]; ptr != NULL; ptr = ptr->next)
      fatmat[(i - 1) * (lp->rows + 1) + ptr->nr.row] = ptr->value;
  
  printf("problem name: %s\n", lp->lp_name);
  printf("          ");
  for(j = 1; j <= lp->columns; j++)
    if(lp->names_used)
      printf("%8s ", lp->col_name[j]);
    else
      printf("Var[%3d] ", j);
  if(lp->maximise)
    {
      printf("\nMaximise  ");
      for(j = 0; j < lp->columns; j++)
	printf("%8g ", (double) -fatmat[j * (lp->rows + 1)]);
    }
  else
    {
      printf("\nMinimize  ");
      for(j = 0; j < lp->columns; j++)
	printf("%8g ", (double) fatmat[j * (lp->rows + 1)]);
    }
  printf("\n");
  for(i = 1; i <= lp->rows; i++) {
    if(lp->names_used)
      printf("%-9s ", lp->row_name[i]);
    else
      printf("Row[%3d]  ", i);
    for(j = 0; j < lp->columns; j++)
      if(lp->ch_sign[i] && fatmat[j*(lp->rows + 1) + i] != 0)
	printf("%8g ", (double)-fatmat[j * (lp->rows+1) + i]);
      else
	printf("%8g ", (double)fatmat[j * (lp->rows + 1) + i]);
    if(lp->orig_upbo[i] != 0) {
      if(lp->ch_sign[i])
	printf(">= ");
      else
	printf("<= ");
    }
    else
      printf(" = ");
    if(lp->ch_sign[i])
      printf("%8g", (double)-lp->orig_rh[i]);
    else
      printf("%8g", (double)lp->orig_rh[i]);
    if(lp->orig_lowbo[i] != 0) {
      printf("  %s = %8g", (lp->ch_sign[i]) ? "lowbo" : "upbo",
	     (double)lp->orig_lowbo[i]);
    }
    if((lp->orig_upbo[i] != lp->infinite) && (lp->orig_upbo[i] != 0.0)) {
      printf("  %s = %8g", (lp->ch_sign[i]) ? "upbo" : "lowbo",
	     (double)lp->orig_upbo[i]);
    }
    printf("\n");
  }
  printf("Type      ");
  for(i = 1; i <= lp->columns; i++)
    if(lp->must_be_int[lp->rows_alloc + i] == TRUE)
      printf("     Int ");
    else
      printf("    Real ");
  printf("\nupbo      ");
  for(i = 1; i <= lp->columns; i++)
    if(lp->orig_upbo[lp->rows_alloc + i] == lp->infinite)
      printf(" Infinite");
    else
      printf("%8g ", (double)lp->orig_upbo[lp->rows_alloc + i]);
  printf("\nlowbo     ");
  for(i = 1; i <= lp->columns; i++)
    printf("%8g ", (double)lp->orig_lowbo[lp->rows_alloc + i]);
  printf("\n");
  for(i = 0; i < lp->nr_lagrange; i++) {
    printf("lag[%d]  ", i);
    for(j = 1; j <= lp->columns; j++)
      printf("%8g ", (double)lp->lag_row[i][j]);
    if(lp->orig_upbo[i] == lp->infinite) {
      if(lp->lag_con_type[i] == GE)
	printf(">= ");
      else if(lp->lag_con_type[i] == LE)
	printf("<= ");
      else if(lp->lag_con_type[i] == EQ)
	printf(" = ");
    }
    printf("%8g\n", (double)lp->lag_rhs[i]);
  }

  free(fatmat);
}  



void set_row_name(lprec *lp, int row, nstring new_name)
{
  int i;
  hashelem *hp;

  if(!lp->names_used) {
    CALLOC(lp->row_name, lp->rows_alloc + 1);
    CALLOC(lp->col_name, lp->columns_alloc + 1);
    lp->names_used = TRUE;
    for(i = 0; i <= lp->rows; i++)
      sprintf(lp->row_name[i], "r_%d", i);
    for(i = 1; i <= lp->columns; i++)
      sprintf(lp->col_name[i], "var_%d", i);
  }
  strcpy(lp->row_name[row], new_name);
  hp = puthash(lp->row_name[row], lp->rowname_hashtab);
  hp->index = row;
}

void set_col_name(lprec *lp, int column, nstring new_name)
{
  int i;
  hashelem *hp;
 
  if(!lp->names_used) {
    CALLOC(lp->row_name, lp->rows_alloc + 1);
    CALLOC(lp->col_name, lp->columns_alloc + 1);
    lp->names_used = TRUE;
    for(i = 0; i <= lp->rows; i++)
      sprintf(lp->row_name[i], "r_%d", i);
    for(i = 1; i <= lp->columns; i++)
      sprintf(lp->col_name[i], "var_%d", i);
  }
  strcpy(lp->col_name[column], new_name);
  hp = puthash(lp->col_name[column], lp->colname_hashtab);
  hp->index = column;
}

static REAL minmax_to_scale(REAL min, REAL max)
{
  REAL scale;

  /* should do something sensible when min or max is 0, MB */
  if((min == 0) || (max == 0))
    return((REAL)1);

  /* scale = 1 / pow(10, (log10(min) + log10(max)) / 2); */
  /* Jon van Reet noticed: can be simplified to: */
  scale = 1 / sqrt(min * max);

  return(scale);
}

void unscale_columns(lprec *lp)
{
  int i, j;
  matrec * ptr;

  /* unscale mat */
  for(j = 1; j <= lp->columns; j++)
    for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
      ptr->value /= lp->scale[lp->rows_alloc + j];

  /* unscale bounds as well */
  for(i = lp->rows_alloc + 1; i <= lp->sum2; i++) { /* was < */ /* changed by PN */
    if(lp->orig_lowbo[i] != 0)
      lp->orig_lowbo[i] *= lp->scale[i];
    if(lp->orig_upbo[i] != lp->infinite)
      lp->orig_upbo[i] *= lp->scale[i];
  }
    
  for(i = lp->rows_alloc + 1; i<= lp->sum2; i++)
    lp->scale[i] = 1;
  lp->columns_scaled = FALSE;
  lp->eta_valid = FALSE;
}


void unscale(lprec *lp)
{
  int i, j;
  matrec * ptr;

  if(lp->scaling_used) {
    /* unscale mat */
    for(j = 1; j <= lp->columns; j++)
      for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
	   ptr->value /= lp->scale[lp->rows_alloc + j];

    /* unscale bounds */
    for(i = lp->rows_alloc + 1; i <= lp->sum2; i++) { /* was < */ /* changed by PN */
      if(lp->orig_lowbo[i] != 0)
	lp->orig_lowbo[i] *= lp->scale[i];
      if(lp->orig_upbo[i] != lp->infinite)
	lp->orig_upbo[i] *= lp->scale[i];
    }
    
    /* unscale the matrix */
    for(j = 1; j <= lp->columns; j++)
      for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
	ptr->value /= lp->scale[ptr->nr.row];
	
    /* unscale the rhs! */
    for(i = 0; i <= lp->rows; i++)
      lp->orig_rh[i] /= lp->scale[i];

    /* and don't forget to unscale the upper and lower bounds ... */
    for(i = 0; i <= lp->rows; i++) {
      if(lp->orig_lowbo[i] != 0)
	lp->orig_lowbo[i] /= lp->scale[i];
      if(lp->orig_upbo[i] != lp->infinite)
	lp->orig_upbo[i] /= lp->scale[i];
    }

    free(lp->scale);
    lp->scaling_used = FALSE;
    lp->eta_valid = FALSE;
  }
}


void auto_scale(lprec *lp)
{
  int i, j, row_nr;
  REAL *row_max, *row_min, *scalechange, absval;
  REAL col_max, col_min;
  matrec * ptr;

  if(!lp->scaling_used) {
    MALLOC(lp->scale, lp->sum_alloc + 1);
    /* for(i = 0; i <= lp->sum2; i++)
       lp->scale[i] = 1;*/
    for(i=0; i<= lp->rows; i++)
      lp->scale[i] = 1;
    for(i=lp->rows_alloc; i<= lp->sum2; i++)
      lp->scale[i] = 1;
  }
  
  MALLOC(row_max, lp->rows + 1);
  MALLOC(row_min, lp->rows + 1);
  MALLOC(scalechange, lp->sum2 + 1);

  /* initialise min and max values */
  for(i = 0; i <= lp->rows; i++) {
    row_max[i] = 0;
    row_min[i] = lp->infinite;
  }

  /* calculate min and max absolute values of rows */
  for(j = 1; j <= lp->columns; j++)
     for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
       {
	  row_nr = ptr->nr.row;
	  absval = my_abs(ptr->value);
	  if(absval != 0) {
	    row_max[row_nr] = my_max(row_max[row_nr], absval);
	    row_min[row_nr] = my_min(row_min[row_nr], absval);
	  }
       }	

  /* calculate scale factors for rows */
  for(i = 0; i <= lp->rows; i++) {
    scalechange[i] = minmax_to_scale(row_min[i], row_max[i]);
    lp->scale[i] *= scalechange[i];
  }

  /* now actually scale the matrix */
  for(j = 1; j <= lp->columns; j++)
    for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
      ptr->value *= scalechange[ptr->nr.row];
 
  /* and scale the rhs and the row bounds (RANGES in MPS!!) */
  for(i = 0; i <= lp->rows; i++) {
    lp->orig_rh[i] *= scalechange[i];
    
    if((lp->orig_upbo[i] < lp->infinite) && (lp->orig_upbo[i] != 0))
      lp->orig_upbo[i] *= scalechange[i];

    if(lp->orig_lowbo[i] != 0)
      lp->orig_lowbo[i] *= scalechange[i];
  }

  free(row_max);
  free(row_min);
  
  /* calculate column scales */
  for(j = 1; j <= lp->columns; j++) {
    if(lp->must_be_int[lp->rows_alloc + j]) { /* do not scale integer columns */
      scalechange[lp->rows_alloc + j] = 1;
    }
    else {
      col_max = 0;
      col_min = lp->infinite;
    	
      for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
	if(ptr->value != 0) {
	  col_max = my_max(col_max, my_abs(ptr->value));
	  col_min = my_min(col_min, my_abs(ptr->value));
	}
      scalechange[lp->rows_alloc + j]  = minmax_to_scale(col_min, col_max);
      lp->scale[lp->rows_alloc + j] *= scalechange[lp->rows_alloc + j];
    }
  }

/* scale mat */
  for(j = 1; j <= lp->columns; j++)
     for(ptr = lp->mat[j]; ptr != NULL; ptr = ptr->next)
     ptr->value *= scalechange[lp->rows_alloc + j];
  
  /* scale bounds as well */
  for(i = lp->rows_alloc + 1; i <= lp->sum2; i++) { /* was < *//* changed by PN */
    if(lp->orig_lowbo[i] != 0)
      lp->orig_lowbo[i] /= scalechange[i];
    if(lp->orig_upbo[i] != lp->infinite)
      lp->orig_upbo[i] /= scalechange[i];
  }
  lp->columns_scaled = TRUE;

  free(scalechange);
  lp->scaling_used = TRUE;
  lp->eta_valid = FALSE;
}

void reset_basis(lprec *lp)
{
  lp->basis_valid = FALSE;
}

void print_solution(lprec *lp)
{
  int i;
  FILE *stream;

  stream = stdout;

  fprintf(stream, "Value of objective function: %g\n",
	  (double)lp->best_solution[0]);

  /* print normal variables */
  for(i = 1; i <= lp->columns; i++)
    if(lp->names_used)
      fprintf(stream, "%-20s %g\n", lp->col_name[i],
	      (double)lp->best_solution[lp->rows_alloc + i]);
    else
      fprintf(stream, "Var [%d] %g\n", i,
	      (double)lp->best_solution[lp->rows_alloc + i]);

  /* print achieved constraint values */
  if(lp->verbose) {
    fprintf(stream, "\nActual values of the constraints:\n");
    for(i = 1; i <= lp->rows; i++)
      if(lp->names_used)
	fprintf(stream, "%-20s %g\n", lp->row_name[i],
		(double)lp->best_solution[i]);
      else
	fprintf(stream, "Row [%d] %g\n", i,
		(double)lp->best_solution[i]);  
  }

  if((lp->verbose || lp->print_duals)) {
    if(lp->max_level != 1)
      fprintf(stream,
	      "These are the duals from the node that gave the optimal solution.\n");
    else
      fprintf(stream, "\nDual values:\n");
    for(i = 1; i <= lp->rows; i++)
      if(lp->names_used)
	fprintf(stream, "%-20s %g\n", lp->row_name[i],
		(double)lp->duals[i]);
      else
	fprintf(stream, "Row [%d] %g\n", i, (double)lp->duals[i]); 
  }
  fflush(stream);
} /* Printsolution */

void write_LP(lprec *lp, FILE *output)
{
  int i, j;
  REAL *row;

  MALLOC(row, lp->columns+1);
  if(lp->maximise)
    fprintf(output, "max:");
  else
    fprintf(output, "min:");

  get_row(lp, 0, row);
  for(i = 1; i <= lp->columns; i++)
    if(row[i] != 0) {
      if(row[i] == -1)
	fprintf(output, " -");
      else if(row[i] == 1)
	fprintf(output, " +");
      else 
	fprintf(output, " %+g ", (double)row[i]);
      if(lp->names_used)
	fprintf(output, "%s", lp->col_name[i]);
      else
	fprintf(output, "x%d", i);
    }
  fprintf(output, ";\n");

  for(j = 1; j <= lp->rows; j++) {
    if(lp->names_used)
      fprintf(output, "%s:", lp->row_name[j]);
    get_row(lp, j, row);
    for(i = 1; i <= lp->columns; i++)
      if(row[i] != 0) {
	if(row[i] == -1)
	  fprintf(output, " -");
	else if(row[i] == 1)
	  fprintf(output, " +");
	else 
	  fprintf(output, " %+g ", (double)row[i]);
	if(lp->names_used)
	  fprintf(output, "%s", lp->col_name[i]);
	else
	  fprintf(output, "x%d", i);
      }
    if(lp->orig_upbo[j] == 0)
      fprintf(output, " =");
    else if(lp->ch_sign[j])
      fprintf(output, " >");
    else
      fprintf(output, " <");
    if(lp->ch_sign[j])
      fprintf(output, " %16g;\n", (double)-lp->orig_rh[j]);
    else
      fprintf(output, " %16g;\n", (double)lp->orig_rh[j]);
  }
  for(i = lp->rows_alloc + 1; i <= lp->sum2; i++) {
    if(lp->orig_lowbo[i] != 0) {
      if(lp->names_used)
	fprintf(output, "%s > %16g;\n", lp->col_name[i - lp->rows_alloc],
		(double)lp->orig_lowbo[i]);
      else
	fprintf(output, "x%d > %16g;\n", i - lp->rows_alloc,
		(double)lp->orig_lowbo[i]);
    }
    if(lp->orig_upbo[i] != lp->infinite) {
      if(lp->names_used)
	fprintf(output, "%s < %16g;\n", lp->col_name[i - lp->rows_alloc],
		(double)lp->orig_upbo[i]);
      else
	fprintf(output, "x%d < %16g;\n", i - lp->rows_alloc,
		(double)lp->orig_upbo[i]);
    }
  }


  i = 1;
  while(!lp->must_be_int[lp->rows_alloc + i]  && i <= lp->columns)
    i++;
  if(i <= lp->columns) {
    if(lp->names_used)  
      fprintf(output, "\nint %s", lp->col_name[i]);
    else
      fprintf(output, "\nint x%d", i);
    i++;
    for(; i <= lp->columns; i++)
      if(lp->must_be_int[lp->rows_alloc + i]) {
	if(lp->names_used)  
	  fprintf(output, ",%s", lp->col_name[i]);
	else
	  fprintf(output, ", x%d", i);
      }
    fprintf(output, ";\n");
  }
  free(row);
}




void write_MPS(lprec *lp, FILE *output)
{
  int i, j, marker, putheader;
  REAL *column, a;


  MALLOC(column, lp->rows + 1);
  marker = 0;   
  fprintf(output, "NAME          %s\n", lp->lp_name);
  fprintf(output, "ROWS\n");
  for(i = 0; i <= lp->rows; i++) {
    if(i == 0)
      fprintf(output, " N  ");
    else
      if(lp->orig_upbo[i] != 0) {
	if(lp->ch_sign[i])
	  fprintf(output, " G  ");
	else
	  fprintf(output, " L  ");
      }
      else
	fprintf(output, " E  ");
    if(lp->names_used)
      fprintf(output, "%s\n", lp->row_name[i]);
    else
      fprintf(output, "r_%d\n", i);
  }
      
  fprintf(output, "COLUMNS\n");

  for(i = 1; i <= lp->columns; i++) {
    if((lp->must_be_int[i + lp->rows_alloc]) && (marker % 2) == 0) {
      fprintf(output,
	      "    MARK%04d  'MARKER'                 'INTORG'\n",
	      marker);
      marker++;
    }
    if((!lp->must_be_int[i + lp->rows_alloc]) && (marker % 2) == 1) {
      fprintf(output,
	      "    MARK%04d  'MARKER'                 'INTEND'\n",
	      marker);
      marker++;
    }
    /* this gets slow for large LP problems. Implement a sparse version? */
    get_column(lp, i, column);
    j = 0;
    if(lp->maximise) {
      if(column[j] != 0) { 
	if(lp->names_used)
	  fprintf(output, "    %-8s  %-8s  %12g\n", lp->col_name[i],
		  lp->row_name[j], (double)-column[j]);
	else
	  fprintf(output, "    var_%-4d  r_%-6d  %12g\n", i, j,
		  (double)-column[j]);
      }
    } 
    else {
      if(column[j] != 0) { 
	if(lp->names_used)
	  fprintf(output, "    %-8s  %-8s  %12g\n", lp->col_name[i],
		  lp->row_name[j], (double)column[j]);
	else
	  fprintf(output, "    var_%-4d  r_%-6d  %12g\n", i, j,
		  (double)column[j]);
      }
    }
    for(j = 1; j <= lp->rows; j++)
      if(column[j] != 0) { 
	if(lp->names_used)
	  fprintf(output, "    %-8s  %-8s  %12g\n", lp->col_name[i],
		  lp->row_name[j], (double)column[j]);
	else
	  fprintf(output, "    var_%-4d  r_%-6d  %12g\n", i, j,
		  (double)column[j]);
      }
  }
  if((marker % 2) == 1) {
    fprintf(output, "    MARK%04d  'MARKER'                 'INTEND'\n",
	    marker);
    /* marker++; */ /* marker not used after this */
  }

  fprintf(output, "RHS\n");
  for(i = 1; i <= lp->rows; i++) {
    a = lp->orig_rh[i];
    if(lp->scaling_used)
      a /= lp->scale[i];

    if(lp->ch_sign[i]) {
      if(lp->names_used)
	fprintf(output, "    RHS       %-8s  %12g\n", lp->row_name[i],
		(double)-a);
      else
	fprintf(output, "    RHS       r_%-6d  %12g\n", i, (double)-a);
    }
    else {
      if(lp->names_used)
	fprintf(output, "    RHS       %-8s  %12g\n", lp->row_name[i],
		(double)a);
      else
	fprintf(output, "    RHS       r_%-6d  %12g\n", i, (double)a);
    }
  }

  putheader = TRUE;
  for(i = 1; i <= lp->rows; i++)
    if((lp->orig_upbo[i] != lp->infinite) && (lp->orig_upbo[i] != 0.0)) {
      if(putheader) {
	fprintf(output, "RANGES\n");
	putheader = FALSE;
      }
      a = lp->orig_upbo[i];
      if(lp->scaling_used)
	a /= lp->scale[i];
      if(lp->names_used)
	fprintf(output, "    RGS       %-8s  %12g\n", lp->row_name[i],
		(double)a);
      else
	fprintf(output, "    RGS       r_%-6d  %12g\n", i,
		(double)a);
    }
    else if((lp->orig_lowbo[i] != 0.0)) {
      if(putheader) {
	fprintf(output, "RANGES\n");
	putheader = FALSE;
      }
      a = lp->orig_lowbo[i];
      if(lp->scaling_used)
	a /= lp->scale[i];
      if(lp->names_used)
	fprintf(output, "    RGS       %-8s  %12g\n", lp->row_name[i],
		(double)-a);
      else
	fprintf(output, "    RGS       r_%-6d  %12g\n", i,
		(double)-a);
    }

  fprintf(output, "BOUNDS\n");
  if(lp->names_used)
    for(i = lp->rows_alloc + 1; i <= lp->sum2; i++) {
      if((lp->orig_lowbo[i] != 0) && (lp->orig_upbo[i] < lp->infinite) &&
	 (lp->orig_lowbo[i] == lp->orig_upbo[i])) {
	a = lp->orig_upbo[i];
	if(lp->scaling_used)
	  a *= lp->scale[i];
	fprintf(output, " FX BND       %-8s  %12g\n",
		lp->col_name[i - lp->rows_alloc], (double)a);
      }
      else {
	if(lp->orig_upbo[i] < lp->infinite) {
	  a = lp->orig_upbo[i];
	  if(lp->scaling_used)
	    a *= lp->scale[i];
	  fprintf(output, " UP BND       %-8s  %12g\n",
		  lp->col_name[i - lp->rows_alloc], (double)a);
	}
	if(lp->orig_lowbo[i] != 0) {
	  a = lp->orig_lowbo[i];
	  if(lp->scaling_used)
	    a *= lp->scale[i];
	  /* bug? should a be used instead of lp->orig_lowbo[i] MB */
	  fprintf(output, " LO BND       %-8s  %12g\n",
		  lp->col_name[i - lp->rows_alloc], (double)lp->orig_lowbo[i]);
	}
      }
    }
  else
    for(i = lp->rows_alloc + 1; i <= lp->sum2; i++) {
      if((lp->orig_lowbo[i] != 0) && (lp->orig_upbo[i] < lp->infinite) &&
	 (lp->orig_lowbo[i] == lp->orig_upbo[i])) {
	a = lp->orig_upbo[i];
	if(lp->scaling_used)
	  a *= lp->scale[i];
	fprintf(output, " FX BND       %-8s  %12g\n",
		lp->col_name[i - lp->rows_alloc], (double)a);
      }
      else {
	if(lp->orig_upbo[i] < lp->infinite) {
	  a = lp->orig_upbo[i];
	  if(lp->scaling_used)
	    a *= lp->scale[i];
	  fprintf(output, " UP BND       var_%-4d  %12g\n",
		  i - lp->rows_alloc, (double)a);
	}
	if(lp->orig_lowbo[i] != 0) {
	  a = lp->orig_lowbo[i];
	  if(lp->scaling_used)
	    a *= lp->scale[i];
	  fprintf(output, " LO BND       var_%-4d  %12g\n", i - lp->rows_alloc,
		  (double)a);
	}
      }
    }
  fprintf(output, "ENDATA\n");
  free(column);
}

void print_duals(lprec *lp)
{
  int i;
  for(i = 1; i <= lp->rows; i++)
    if(lp->names_used)
      fprintf(stdout, "%s [%d] %g\n", lp->row_name[i], i,
	      (double)lp->duals[i]);
    else
      fprintf(stdout, "Dual [%d] %g\n", i, (double)lp->duals[i]);
}

void print_scales(lprec *lp)
{
  int i;
  if(lp->scaling_used) {
    for(i = 0; i <= lp->rows; i++)
      fprintf(stdout, "Row[%d]    scaled at %g\n", i,
	      (double)lp->scale[i]);
    for(i = 1; i <= lp->columns; i++)
      fprintf(stdout, "Column[%d] scaled at %g\n", i,
	      (double)lp->scale[lp->rows_alloc + i]);
  }
}


void free_mat_rec(matrec * elem)
{
  if (elem != NULL)
    {
      free_mat_rec(elem->next);
      free(elem);
    }
  elem = NULL;
}

matrec * copy_mat_rec(matrec * elem)
{
  if (elem == NULL)
    return NULL;
  else
    {
      matrec * nelem;
      MALLOCCPY(nelem, elem, 1);
      nelem->next = copy_mat_rec(elem->next);
      return nelem;
    }
} 


