/*-------------------------------------------------------------------------*
 * MEM DEBUG                                                               *
 *                                                                         *
 * File  : mem_debug.c                                                     *
 * Descr.: a helpfull library for memory allocation debugging              *
 * Author: Remy Haemmerle <Remy.Haemmerle@inria.fr>                        *
 *                                                                         *
 * Copyright (C) 2002 Remy Haemmerle                                       *
 *                                                                         *
 * GNU Prolog is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2, or any later version.       *
 *                                                                         *
 * GNU Prolog is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc.  *
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     *
 *-------------------------------------------------------------------------*/

/* Version 0.2 - August 2002 */

#define _mem_debug_c

#include "mem_debug.h"

#ifdef   NO_MEM_DEBUG

void * 
calloc_d(int n, int s)
{
  return calloc(n, s);
}

void * 
malloc_d(int n)
{
  return malloc(n);
}

void 
free_d(void * ptr)
{
  free(ptr);
}

void * 
realloc_d(void * ptr, int s)
{
  return realloc(ptr, s);
}

#else /* NO_MEM_DEBUG */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*---------------------------------*
 * Constants                       *
 *---------------------------------*/

#define min(_a, _b)     ((_a < _b) ? _a : _b) 

#define MOTIF_DIFF(_ptr, _nb, _motif, _i)         \
    for(_i=0; _i<_nb; _i++)                       \
      {                                           \
        if ( _ptr[_i] !=  _motif)                 \
          {                                       \
	    memd_error("pb memory");            \
	  }                                       \
      }                                           


#if (CHECK_EVERY == 0)
#define CHECK_MEM()
#else
#define CHECK_MEM()                               \
    if (counter1 > CHECK_START)                   \
      {                                           \
	if (check != 0)                           \
	  {                                       \
	    if (counter2 == CHECK_EVERY)          \
	      {                                   \
	        counter2 = 1;                     \
	        verify_mem();                     \
              }                                   \
            else                                  \
              {                                   \
                counter2 ++;                      \
              }                                   \
          }                                       \
      }                                           \
    counter1++;                    
#endif /* (CHECK_EVERY == 0) */


/*---------------------------------*/
/* Global Variables                */
/*---------------------------------*/

static short  memd_status = 0;
static unsigned char   main_mem[SIZE_MEM];
static unsigned char   mem_state[NBR_OF_BLOCKS];
static int mem_counter = 0;

static int counter1 = 0;
static int counter2 = 1;
static int check = (CHECK_START != 0);


/*---------------------------------*/
/* Function Prototypes             */
/*---------------------------------*/

void memd_error( char * message);
void memd_init();
unsigned char * memd_alloc(int s, unsigned char motif);
void verify_mem();

void 
memd_error(char * msg)
{
  printf("%s   (%i)\n",msg, counter1);
  exit(1);
}

void
memd_init()
{
  int i;

#ifdef PRINT_INFO
  printf("memd_init : preparing memory :");
  fflush(stdout);
#endif

  for(i=0; i < SIZE_MEM; i++)
    {
      main_mem[i] =  NALLOC_MOTIF;
    }
  for(i=0; i < NBR_OF_BLOCKS; i++)
    {
      mem_state[i] = NON_ALLOCATED;
    }

  memd_status = 1;

#ifdef PRINT_INFO
  printf(" OK\n");
#endif
}


unsigned char *
memd_alloc(int size, unsigned char motif)
{
  int i;
  int new_counter;
  unsigned char * ptr;

  if (memd_status == 0) 
    {
      memd_init();
    }

  mem_counter += MEM_SKIP;

  new_counter = ( size + SIZE_BLOCK - 1) / SIZE_BLOCK + mem_counter;

  if (new_counter > NBR_OF_BLOCKS)
    {
      memd_error("memd_alloc : memory full");
    }
    
  ptr = main_mem + mem_counter * SIZE_BLOCK;

  mem_state[mem_counter] = ALLOCATED;
  mem_counter++;

  for(; mem_counter < new_counter; mem_counter++)
    {
      mem_state[mem_counter] = OWNED;
    }

  for(i=0; i<size; i++)
    {
      ptr[i] = motif;
    }

  /*  CHECK_MEM();*/

  return ptr;
}


void * 
calloc_d(int n, int s)
{
  return ((void *) memd_alloc(n * s, BLANK_MOTIF));
}


void *
malloc_d(int s)
{
  return ((void *) memd_alloc(s, ALLOC_MOTIF));
}


void 
free_d(void * ptr)
{
  int counter;
  int i,j;

  if (ptr == NULL) return;

  i = (int) (((unsigned char *) ptr) - main_mem);

  counter = i / SIZE_BLOCK;

  if ((i - counter * SIZE_BLOCK) != 0)
    {
      memd_error("free_d : pointer doesn\'t correspond to a block");
    }

  if ((counter < 0) || (counter > NBR_OF_BLOCKS))
    {
      memd_error("free_d : pointer not in main_mem");
    }

  switch (mem_state[counter])
    {
    case NON_ALLOCATED:
      memd_error("free_d : pointer never allocated");
    case ALLOCATED:
      break;
    case OWNED:
      memd_error("free_d : invalid pointer");
    case FREE :
      memd_error("free_d : pointer already desallocated");
    default:
      memd_error("free_d : mem_state corrupted (1)");
    }

  i = counter;
  mem_state[i] = OWNED;
  
  while(mem_state[i] == OWNED)
    {
      mem_state[i] = FREE;
      for(j=0; j < SIZE_BLOCK; j++)
	{
	  main_mem[i*SIZE_BLOCK+j] = FREE_MOTIF;
	}
      i++;
    }


#if (MEM_SKIP == 0)
  return;
#else
  if (mem_state[i] == NON_ALLOCATED)
    {
      /*     CHECK_MEM();*/
      return;
    }
  else  
    {
      memd_error("free_d : mem_state corrupted (2)");
    }
#endif


}

void *
realloc_d(void * ptr, int size)
{
  int counter, nb;
  int i;
  unsigned char * tmp;


  if (size == 0) 
    {
      free_d(ptr);
      return NULL;
    }
  
  if (ptr == NULL)
    {
      return malloc_d(size);
    }

  i = (int) (((unsigned char *) ptr) - main_mem);

  counter = i / SIZE_BLOCK;

  if ((i - counter * SIZE_BLOCK) != 0)
    {
      memd_error("realloc_d : pointer doesn\'t correspond to a block");
    }

  if ((counter < 0) || (counter > NBR_OF_BLOCKS))
    {
      memd_error("realloc_d : pointer out of  main_mem");
    }

  switch (mem_state[counter])
    {
    case NON_ALLOCATED:
      memd_error("realloc_d : pointer never allocated");
    case ALLOCATED:
      break;
    case OWNED:
      memd_error("realloc_d : invalid pointer");
    case FREE :
      memd_error("realloc_d : pointer desallocated");
    default:
      memd_error("realloc_d : mem_state corrupted (1)");
    }

  i = counter + 1;

  while(mem_state[i] == OWNED)
    {
      i++;
    }

  nb = i - counter;
  
  tmp = memd_alloc(size, ALLOC_MOTIF);
  
  memcpy(tmp, ptr, min(nb * SIZE_BLOCK, size));
  
  free_d(ptr);

  return (void *) tmp;
}

void
verify_mem(void)
{
  int i, j;
  unsigned char *ptr;
  
  int  allocated = 0;

  ptr = main_mem;

  printf("checking memory :");
  fflush(stdout);

  for(i=0; i<NBR_OF_BLOCKS; i++)
    {
      switch(mem_state[i])
	{
	case NON_ALLOCATED:
	  MOTIF_DIFF(ptr, SIZE_BLOCK, NALLOC_MOTIF, j);
	  allocated = 0;
	  break;
	case ALLOCATED:
#if (MEM_SKIP != 0)
	  if (allocated != 0)
	    {
	      memd_error("memory problem : no nomansland\n");
	    }
#endif  
	  allocated = 1;
	  break;
	case OWNED:
	  if (allocated == 0)
	    {
	      memd_error("memory problem : owned without allocated before\n");
	    }
	  allocated = 2;
	  break;
	case FREE:
	  MOTIF_DIFF(ptr, SIZE_BLOCK, FREE_MOTIF, j);
	  allocated = 0;
	}
      ptr += SIZE_BLOCK;
    }

  printf(" OK\n");

}

void 
check_mem()
{
  CHECK_MEM();
}

#endif    /* NO_MEM_DEBUG */
