/*-------------------------------------------------------------------------*
 * MEM DEBUG                                                               *
 *                                                                         *
 * File  : mem_debug.h                                                     *
 * Descr.: a helpfull library for memory allocation debugging (header)     *
 * Author: Remy Haemmerle <Remy.Haemmerle@inria.fr>                        *
 *                                                                         *
 * Copyright (C) 2002 Remy Haemmerle                                       *
 *                                                                         *
 * GNU Prolog is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2, or any later version.       *
 *                                                                         *
 * GNU Prolog is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc.  *
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     *
 *-------------------------------------------------------------------------*/

/* Version 0.2 - August 2002 */

#ifndef _mem_debug_h
#define _mem_debug_h 

#include "memd_config.h"

/*---------------------------------*
 * Constants                       *
 *---------------------------------*/

#define KILO_BYTE       1024
#define MEGA_BYTE       1024 *  KILO_BYTE

#define NALLOC_MOTIF    0x95
#define BLANK_MOTIF     0x00
#define FREE_MOTIF      0x59
#define ALLOC_MOTIF     0x36

#define NON_ALLOCATED   0x01
#define ALLOCATED       0x02
#define OWNED           0x04
#define FREE            0X08

#define PBR             0X10
#define NULL_PTR        0x20
#define BAD_PTR         0x40

#define NBR_OF_BLOCKS   SIZE_MEM / SIZE_BLOCK
#define MEM_SKIP        NOMANSLAND / SIZE_BLOCK


#ifndef _mem_debug_c
#ifndef NO_MEM_DEBUG

#define free(ptr)        free_d(ptr)
#define malloc(s)        malloc_d(s)
#define calloc(s, n)     calloc_d(s, n)
#define realloc(ptr, s)  realloc_d(ptr, s)

#endif
#endif

/*---------------------------------*/
/* Global Variables                */
/*---------------------------------*/


/*---------------------------------*/
/* Function Prototypes             */
/*---------------------------------*/

void * calloc_d(int n, int s);
void * malloc_d(int n);
void free_d(void * ptr);
void * realloc_d(void * ptr, int s);

char state_ptr(void * ptr);
void check_mem();

#endif /* _mem_debug_h */
